"""
Dynamic nonlinear period-by-period simulator
"""


#[
from __future__ import annotations

from numbers import (Real, )
from typing import (Callable, Any, Literal, )
import numpy as _np
import scipy as _sp
import functools as _ft

from ..evaluators.base import (DEFAULT_INIT_GUESS, )
from ..simultaneous import main as _simultaneous
from ..plans import main as _plans
from .. import quantities as _quantities
from .. import equations as _equations
from .. import dataslates as _dataslates
from .. import wrongdoings as _wrongdoings
from . import evaluators as _evaluators
#]


_DEFAULT_FALLBACK_VALUE = 1/9


def simulate(
    model: _simultaneous.Simultaneous,
    dataslate: _dataslate.Dataslate,
    vid: int,
    /,
    *,
    plan: _plans.PlanSimulate | None,
    start_iter_from: Literal["data", "first_order", "previous_period"] = "previous_period",
    root_settings: dict[str, Any] | None = None,
    iter_printer_settings: dict[str, Any] | None = None,
    when_fails: Literal["critical", "error", "warning", "silent"] = "critical",
    when_missing: Literal["critical", "error", "warning", "silent"] = "critical",
    fallback_value: Real = _DEFAULT_FALLBACK_VALUE,
) -> None:
    """
    """
    root_settings = (
        _DEFAULT_ROOT_SETTINGS
        if root_settings is None
        else _DEFAULT_ROOT_SETTINGS | root_settings
    )

    qid_to_name = model.create_qid_to_name()
    name_to_qid = model.create_name_to_qid()
    all_quantities = model.get_quantities()
    endogenous_names = model.simulate_can_be_exogenized
    exogenous_names = model.simulate_can_be_endogenized
    endogenous_qids = tuple(name_to_qid[name] for name in endogenous_names)
    exogenous_qids = tuple(name_to_qid[name] for name in exogenous_names)
    wrt_qids = endogenous_qids
    wrt_equations = model.get_dynamic_equations(kind=_equations.EquationKind.ENDOGENOUS_EQUATION, )

    if len(wrt_equations) != len(wrt_qids):
        raise _wrongdoings.IrisPieError(
            f"Number of endogenous equations {len(wrt_equations)} "
            f"does not match number of endogenous quantities {len(wrt_qids)}"
        )

    if len(wrt_equations) == 0:
        return

    evaluator_factory = _ft.partial(
        _evaluators.PeriodEvaluator,
        wrt_equations=wrt_equations,
        all_quantities=all_quantities,
        context=model.get_context(),
        iter_printer_settings=iter_printer_settings,
    )

    base_evaluator = evaluator_factory(wrt_qids, )

    when_missing_stream = \
        _wrongdoings.STREAM_FACTORY[when_missing] \
        (f"These values are missing at the start of simulation: ")

    starter = _ITER_STARTER[start_iter_from]

    catch_missing = _ft.partial(
        _catch_missing,
        dataslate=dataslate,
        qid_to_name=qid_to_name,
        fallback_value=fallback_value,
        when_missing_stream=when_missing_stream,
    )

    for t in dataslate.base_columns:
        current_wrt_qids, current_evaluator = \
            _set_up_current_period(plan, evaluator_factory, wrt_qids, dataslate.column_dates[t], base_evaluator, name_to_qid, )
        current_evaluator.iter_printer.header_message = \
            _create_header_message(vid, t, dataslate.column_dates[t], )
        dataslate.data[current_wrt_qids, t] = starter(dataslate.data, current_wrt_qids, t, )
        catch_missing(dataslate.data, t, )
        #
        init = current_evaluator.get_init_guess(dataslate.data, t, )
        root_final = _sp.optimize.root(
            current_evaluator.eval, init,
            args=(dataslate.data, t, None, ),
            jac=True,
            **root_settings,
        )
        #
        if not root_final.success:
            _catch_fail(root_final, when_fails, )
        current_evaluator.update(root_final.x, dataslate.data, t, )
        #
        current_evaluator.iter_printer.print_footer()
        current_evaluator.iter_printer.reset()


# REFACTOR
def _set_up_current_period(
    plan: _plans.PlanSimulate | None,
    evaluator_factory: Callable[..., _evaluators.PeriodEvaluator],
    wrt_qids: tuple[int, ...],
    current_period: _dates.Dater,
    base_evaluator: _evaluators.PeriodEvaluator,
    name_to_qid: dict[str, int],
    /,
) -> tuple[tuple[int, ...], _evaluators.PeriodEvaluator]:
    """
    """
    if plan is None:
        return wrt_qids, base_evaluator
    #
    names_exogenized = plan.get_names_exogenized_in_period(current_period, )
    names_endogenized = plan.get_names_endogenized_in_period(current_period, )
    if not names_exogenized and not names_endogenized:
        return wrt_qids, base_evaluator
    #
    if len(names_exogenized) != len(names_endogenized):
        raise _wrongdoings.IrisPieCritical(
            f"Number of exogenized quantities {len(names_exogenized)}"
            f" does not match number of endogenized quantities {len(names_endogenized)}"
            f" in period {current_period}"
        )
    #
    qids_exogenized = tuple(name_to_qid[name] for name in names_exogenized)
    qids_endogenized = tuple(name_to_qid[name] for name in names_endogenized)
    current_wrt_qids = tuple(sorted(set(wrt_qids).difference(qids_exogenized).union(qids_endogenized)))
    current_evaluator = evaluator_factory(current_wrt_qids, )
    return current_wrt_qids, current_evaluator


def _start_iter_from_previous_period(
    data: _np.ndarray,
    wrt_qids: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    source_t = t - 1 if t > 0 else 0
    return data[wrt_qids, source_t]


def _start_iter_from_data(
    data: _np.ndarray,
    wrt_qids: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    return data[wrt_qids, t]


def _start_iter_from_first_order(
    data: _np.ndarray,
    wrt_qid: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    raise NotImplementedError()


def _catch_missing(
    data: _np.ndarray,
    t: int,
    /,
    dataslate: _dataslate.Dataslate,
    qid_to_name: dict[int, str],
    fallback_value: Real,
    when_missing_stream: _wrongdoings.Stream,
) -> None:
    """
    """
    missing = _np.isnan(data[:, t])
    if not missing.any():
        return
    #
    data[missing, t] = fallback_value
    #
    current_period = dataslate.column_dates[t]
    shift = 0
    for qid in _np.flatnonzero(missing):
        when_missing_stream.add(
            f"{qid_to_name[qid]}[{current_period+shift}]"
            f" when simulating {current_period}"
        )


def _catch_fail(
    root_final: _sp.optimize.OptimizeResult,
    when_fails: Literal["error", "warning", "silent"],
) -> None:
    """
    """
    raise ValueError()


def _create_header_message(
    vid: str,
    t: int,
    current_period: int,
) -> str:
    """
    """
    return f"[Variant {vid}][Period {current_period}]"


_ITER_STARTER = {
    "previous_period": _start_iter_from_previous_period,
    "data": _start_iter_from_data,
    "first_order": _start_iter_from_first_order,
}


_DEFAULT_ROOT_SETTINGS = {
    "method": "hybr",
    "tol": 1e-12,
}




"""
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
import parsimonious as _pa
import functools as _ft
import re as _re

from . import common as _common
from . import _substitutions as _substitutions
#]


_WHERE_SUBSTITUTE = ["transition-equations", "measurement-equations"]


def from_string(source_string: str, /, ) -> sources.Source:
    """
    """
    source_string = _resolve_shortcut_keywords(source_string)
    source_string = _common.add_blank_lines(source_string)
    source_string = _translate_keywords(source_string)
    nodes = _GRAMMAR["source"].parse(source_string)
    parsed = _Visitor().visit(nodes)
    parsed = _substitutions.resolve_substitutions(parsed, _WHERE_SUBSTITUTE)
    return parsed


_GRAMMAR_DEF = _common.GRAMMAR_DEF + r"""

    source = block+
    block = white_spaces (qty_block / log_block / eqn_block) white_spaces

    keyword = 
        transition_equations_keyword / measurement_equations_keyword
        / transition_variables_keyword / transition_shocks_keyword / measurement_variables_keyword / measurement_shocks_keyword / parameters_keyword / exogenous_variables_keyword
        / log_keyword / all_but_keyword
        / substitutions_keyword / autoswaps_simulate_keyword / autoswaps_steady_keyword
        / preprocessor_keyword / postprocessor_keyword

    description = (description_quote description_text description_quote)?
    description_quote = '"'
    description_text = ~r'[^"]*'

    block_attributes = "{" attribute_chain "}"
        attribute_chain = ~r"(\s*:\w+\s*)+"

    eqn_block = eqn_keyword block_attributes? eqn_ended* 
    eqn_keyword = transition_equations_keyword / measurement_equations_keyword
        / substitutions_keyword / autoswaps_simulate_keyword / autoswaps_steady_keyword
        / preprocessor_keyword / postprocessor_keyword
    eqn_ended = white_spaces description white_spaces eqn_body eqn_end
    eqn_end = ";" 

    eqn_body = eqn_version eqn_steady
    eqn_steady = (eqn_version_separator eqn_version)?
    eqn_version_separator = "!!"
    eqn_version = ~r"[\?\w\.,\(\)=:\+\^\-\*\{\}\[\]/ \t\n&\$]+"

    qty_block = qty_keyword block_attributes? qty_ended*
    qty_keyword = transition_variables_keyword / transition_shocks_keyword / measurement_variables_keyword / measurement_shocks_keyword / parameters_keyword / exogenous_variables_keyword
    qty_ended = white_spaces description white_spaces qty_name qty_end
    qty_name = ~r"\b[\?a-zA-Z]\w*\b"
    qty_end = ~r"[,;\s\n]+"

    log_block = log_keyword white_spaces all_but_flag white_spaces log_ended*
    all_but_flag = (all_but_keyword)?
    log_ended = white_spaces qty_name qty_end

    transition_variables_keyword = keyword_prefix "transition-variables"
    transition_shocks_keyword = keyword_prefix "transition-shocks"
    measurement_variables_keyword = keyword_prefix "measurement-variables"
    measurement_shocks_keyword = keyword_prefix "measurement-shocks"
    parameters_keyword = keyword_prefix "parameters"
    exogenous_variables_keyword = keyword_prefix "exogenous-variables"
    transition_equations_keyword = keyword_prefix "transition-equations"
    equations_keyword = keyword_prefix "equations"
    measurement_equations_keyword = keyword_prefix "measurement-equations"
    log_keyword = keyword_prefix "log-variables"
    all_but_keyword = keyword_prefix "all-but"
    substitutions_keyword = keyword_prefix "substitutions"
    autoswaps_simulate_keyword = keyword_prefix "autoswaps-simulate"
    autoswaps_steady_keyword = keyword_prefix "autoswaps-steady"
    preprocessor_keyword = keyword_prefix "preprocessor"
    postprocessor_keyword = keyword_prefix "postprocessor"

"""


_GRAMMAR = _pa.grammar.Grammar(_GRAMMAR_DEF)
_KEYWORDS = [ k.members[1].literal for k in _GRAMMAR["keyword"].members ]
_KEYWORDS_PATTERN = _common.compile_keywords_pattern(_KEYWORDS)
_translate_keywords = _ft.partial(_common.translate_keywords, _KEYWORDS_PATTERN)


class _Visitor(_pa.nodes.NodeVisitor):
    """
    """
    #[
    def __init__(self):
        super().__init__()
        self.content = {}

    def _add(self, block_name, new_content, ):
        if new_content:
            updated_content = self.content.setdefault(block_name, []) + new_content
            self.content[block_name] = updated_content

    def visit_source(self, node, visited_children):
        return self.content

    def _visit_block(self, node, visited_children):
        if not visited_children[2]:
            return
        block_name = visited_children[0]
        block_attributes = \
            (visited_children[1][0], ) \
            if visited_children[1] else (None, )
        block_content_with_attributes = [
            b + block_attributes
            for b in visited_children[2]
        ]
        self._add(block_name, block_content_with_attributes, )

    visit_qty_block = _visit_block
    visit_eqn_block = _visit_block

    def visit_block_attributes(self, node, visited_children):
        attributes = visited_children[1].strip().replace("(", "").replace(")", "")
        attributes = attributes.split(":")
        if len(attributes) < 2:
            return ()
        attributes = tuple(":" + a.strip() for a in attributes[1:])
        return attributes

    def visit_log_block(self, node, visited_children):
        block_name, _, _, _, block_content = visited_children
        self._add(block_name, block_content, )

    def visit_all_but_flag(self, node, visited_children):
        block_name = "all-but"
        flag = visited_children[0] if visited_children else ""
        self._add(block_name, [flag], )
        return flag

    def visit_qty_ended(self, node, visited_children):
        _, description, _, name, _ = visited_children
        return (description, name)

    def visit_eqn_ended(self, node, visited_children):
        _, description, _, eqn, _ = visited_children
        return (description, eqn)

    def visit_log_ended(self, node, visited_children):
        _, name, _ = visited_children
        return name

    def _visit_keyword(self, node, visited_children):
        return visited_children[1]

    def _visit_abbreviated_keyword(self, node, visited_children):
        return "transition-" + visited_children[1]

    visit_transition_variables_keyword = _visit_keyword
    visit_transition_shocks_keyword = _visit_keyword
    visit_measurement_variables_keyword = _visit_keyword
    visit_measurement_shocks_keyword = _visit_keyword
    visit_parameters_keyword = _visit_keyword
    visit_exogenous_variables_keyword = _visit_keyword
    visit_transition_equations_keyword = _visit_keyword
    visit_equations_keyword = _visit_abbreviated_keyword
    visit_measurement_equations_keyword = _visit_keyword
    visit_substitutions_keyword = _visit_keyword
    visit_autoswaps_simulate_keyword = _visit_keyword
    visit_autoswaps_steady_keyword = _visit_keyword
    visit_preprocessor_keyword = _visit_keyword
    visit_postprocessor_keyword = _visit_keyword
    visit_log_keyword = _visit_keyword
    visit_all_but_keyword = _visit_keyword

    def visit_qty_keyword(self, node, visited_children):
        return visited_children[0]

    def visit_eqn_keyword(self, node, visited_children):
        return visited_children[0]

    def visit_qty_name(self, node, visited_children):
        return node.text.strip()

    def visit_eqn_body(self, node, visited_children):
        dynamic = _deblank(visited_children[0])
        steady = _deblank(visited_children[1])
        return (dynamic, steady)

    def visit_eqn_steady(self, node, visited_children):
        return visited_children[0][1] if visited_children else ""

    def visit_description(self, node, visited_children):
        return visited_children[0][1] if visited_children else ""

    def visit_eqn_version(self, node, visited_children):
        return node.text.strip()

    def visit_description_text(self, node, visited_children):
        return node.text

    def generic_visit(self, node, visited_children):
        return visited_children or node.text or None
    #]


_SHORTCUT_KEYWORDS = [
    ( _re.compile(_common.HUMAN_PREFIX + r"variables\b"), _common.HUMAN_PREFIX + r"transition-variables" ),
    ( _re.compile(_common.HUMAN_PREFIX + r"shocks\b"), _common.HUMAN_PREFIX + r"transition-shocks" ),
    ( _re.compile(_common.HUMAN_PREFIX + r"equations\b"), _common.HUMAN_PREFIX + r"transition-equations" ),
]


def _resolve_shortcut_keywords(source_string: str, /, ) -> str:
    for short, long in _SHORTCUT_KEYWORDS:
        source_string = _re.sub(short, long, source_string)
    return source_string


def _deblank(string: str, /, ) -> str:
    return "".join(string.split())




"""
"""


#[
from __future__ import annotations

from typing import (Self, Any, )
import prettytable as _pt
#]


class PrettyMixin:
    @property
    def pretty(self, /, ) -> _pt.PrettyTable:
        """
        """
        return self.get_pretty()

    @property
    def pretty_full(self, /, ) -> str:
        """
        """
        return self.get_pretty(full=True, )

    def get_pretty(
        self,
        /,
    ) -> _pt.PrettyTable:
        """
        """
        table = _pt.PrettyTable()
        table.field_names = ("NAME", "PERIOD", "REGISTER", "TRANSFORM", )
        table.align = "r"
        table.align["NAME"] = "l"
        for r in ("exogenized", "endogenized", "anticipated", ):
            if getattr(self, f"_{r}_register"):
                _add_register_to_table(
                    table,
                    self.base_span,
                    getattr(self, f"_{r}_register"),
                    r,
                )
        return table

    def pretty_print(self, *args, **kwargs, ) -> None:
        """
        """
        print(self.get_pretty_table(*args, **kwargs, ), )

    def get_pretty_string(self, *args, **kwargs, ) -> str:
        """
        """
        return self.get_pretty_table(*args, **kwargs, ).get_string()


def _add_register_to_table(
    table,
    base_span,
    register,
    action,
) -> None:
    """
    """
    #[
    for k, v in register.items():
        for status, date in zip(v, base_span):
            if not status:
                continue
            symbol = status.symbol if hasattr(status, "symbol") else _PRETTY_SYMBOL.get(status, "")
            table.add_row((k, str(date), action, symbol, ))
    #]


_PRETTY_SYMBOL = {
    None: "",
    True: "*",
    False: "",
}




"""
"""


#[
from __future__ import annotations

from typing import (Self, )
from numbers import (Number, )
import numpy as _np

from ..incidences import main as _incidence
from . import differentiators as ad_
from . import _rules as _rules

#]


class Atom(ad_.LoglyMixin):
    """
    Atomic value for invariance testing
    """
    _data_context: _np.ndarray | None = None
    _logly_context: dict[int, bool] | None = None
    _is_atom: bool = True
    #[
    def __init__(self) -> None:
        """
        """
        self._diff = None
        self._invariant = None
        self._logly = None
        self._logly_index = None

    @classmethod
    def no_context(
        cls: type,
        diff: _np.ndarray,
        invariant: _np.ndarray,
        logly: bool,
    ) -> Self:
        self = cls()
        self._diff = diff
        self._invariant = invariant
        return self

    @classmethod
    def in_context(
        cls: type,
        diff: _np.ndarray,
        token: _incidence.Token,
        *args,
    ) -> Self:
        """
        """
        self = cls()
        self._diff = diff==1
        self._invariant = _np.full((diff.shape[0], 1), True)
        self._logly_index = token.qid
        return self

    __add__ = _rules.add_invar

    __sub__ = _rules.sub_invar

    def __mul__(self, other: Self | Number):
        """
        Invariance of self(x)*other(x) or self(x)*other
        """
        if hasattr(other, "_is_atom"):
            new_diff = self._diff | other._diff
            if _np.all(self._diff==False):
                new_invariant = other._invariant
                new_logly = other.logly
            elif _np.all(other._diff==False):
                new_invariant = self._invariant
                new_logly = self.logly
            else:
                new_invariant = (
                    _np.logical_not(self._diff) & other._invariant
                    & _np.logical_not(other._diff) & self._invariant
                )
                new_logly = False
        else:
            new_diff = self._diff
            new_invariant = self._invariant
            new_logly = self.logly
        return Atom.no_context(new_diff, new_invariant, new_logly)

    def __rmul__(self, other):
        """
        Invariance of other*self(x)
        """
        return Atom.no_context(self._diff, self._invariant, self._logly)

    def __truediv__(self, other):
        """
        Invariance of self(x)/other(x) or self(x)/other
        """
        if hasattr(other, "_is_atom"):
            new_diff = self._diff | other._diff
            new_invariant = _np.all(other._diff, axis=0) & self._invariant
            new_logly = False
        else:
            new_diff = self._diff
            new_invariant = self._invariant
            new_logly = self._logly
        return Atom.no_context(new_diff, new_invariant, new_logly)

    def __rtruediv__(self, other: Number) -> Self:
        """
        Invariance of self(x) / other
        """
        new_diff = self._diff
        new_invariant = False & self._invariant
        return Atom.no_context(new_diff, new_invariant, False)

    __radd__ = __add__

    __rsub__ = __add__

    def _log_(self) -> Self:
        """
        Invariance of log(self(x))
        """
        new_diff = self._diff
        new_invariant = new_invariant if self._logly else (False & self._invariant)
        return Atom.no_context(new_diff, new_invariant, False)

    def _unary(self) -> Self:
        """
        Invariance of f(self(x))
        """
        new_diff = self._diff
        new_invariant = False & self._invariant
        return Atom.no_context(new_diff, new_invariant, False)

    def _binary(self, other: Self | Number) -> Self:
        """
        Invariance of f(self(x), other(x)) or f(self(x), other)
        """
        if hasattr(other, "_is_atom"):
            new_diff = self._diff | other._diff
        else:
            new_diff = self._diff
        new_invariant = False & self._invariant
        return Atom.no_context(new_diff, new_invariant, False)

    exp = _unary
    __pow__ = _binary
    #]





"""
"""


#[
from __future__ import absolute_import

from typing import (Sequence, Iterable, )
import plotly.graph_objects as _pg
import plotly.subplots as _ps

from . import dates as _dates
#]


__all__ = (
    "make_subplots",
    "highlight",
)


_EMPTY_SUBPLOT_TITLE = " "


def make_subplots(
    rows: int,
    columns: int | None = None,
    cols: int | None = None,
    subplot_titles: Sequence[str] | bool | None = True,
    vertical_spacing: float | None = 0.1,
    horizontal_spacing: float | None = 0.05,
    **kwargs,
) -> _pg.Figure:
    """
    """
    if cols is not None and columns is None:
        columns = cols
    total_num_subplots = rows * columns
    if subplot_titles is True:
        subplot_titles = (_EMPTY_SUBPLOT_TITLE, )*total_num_subplots
    return _ps.make_subplots(
        rows=rows,
        cols=columns,
        subplot_titles=subplot_titles or None,
        vertical_spacing=vertical_spacing,
        horizontal_spacing=horizontal_spacing,
        **kwargs,
    )


def highlight(
    figure: _pg.Figure,
    span: Iterable[_dates.Dater],
    /,
    subplot: tuple[int, int] | int | None = None,
    fill_color: str = "rgba(0, 0, 0, 0.15)",
) -> _pg.Figure:
    """
    """
    span = tuple(span)
    _, index = resolve_subplot(figure, subplot, )
    xref = f"x{index+1}" if index else "x"
    yref = f"y{index+1} domain" if index else "y domain"
    shape = {
        "type": "rect",
        "xref": xref,
        "x0": span[0].to_plotly_date(position="start", ),
        "x1": span[-1].to_plotly_date(position="end", ),
        "yref": yref,
        "y0": 0,
        "y1": 1,
        "fillcolor": fill_color,
        "line": {"width": 0, },
    }
    figure.add_shape(shape, )


def resolve_subplot(
    figure: _pg.Figure,
    subplot: tuple[int, int] | int | None,
    /,
) -> tuple[tuple[int, int], int]:
    """
    """
    if subplot is None:
        tile = None
        index = None
        return tile, index,
    rows, columns = figure._get_subplot_rows_columns()
    num_rows = len(rows)
    num_columns = len(columns)
    if isinstance(subplot, Sequence):
        row = subplot[0]
        column = subplot[1]
        index = row * num_columns + column
        tile = row, column,
        return tile, index,
    if isinstance(subplot, int):
        index = subplot
        row = index // num_columns
        column = index % num_columns
        tile = row, column,
        return tile, index,
    raise TypeError(f"Invalid subplot type: {type(subplot)}")




"""
"""


#[
from __future__ import annotations

from .. import quantities as _quantities
from ..databoxes import main as _databoxes
#]


class AssignMixin:
    """
    """
    #[

    def assign(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
        """
        for v in self._variants:
            v.assign(*args, **kwargs, )

    def get_parameters(self, /, ) -> _databoxes.Databox:
        """
        """
        parameter_names = self._invariant.parameter_names
        if self.is_singleton:
            return _databoxes.Databox(**{
                n: self._variants[0].parameters[n]
                for n in parameter_names
            })
        else:
            return _databoxes.Databox(**{
                n: [ v.parameters[n] for v in self._variants ]
                for n in parameter_names
            })

    #]




"""
First-order system simulators
"""

#[
from __future__ import annotations

from typing import (Self, TypeAlias, Literal, )
import numpy as _np

from .. import dataslates as _dataslates
from ..plans import main as _plans
from ..fords import solutions as _solutions
from ..fords import descriptors as _descriptors
#]


def simulate_flat(
    solution: _solutions.Solution,
    solution_vectors: _descriptors.SolutionVectors,
    dataslate: _dataslates.HorizontalDataslate,
    vid: int,
    /,
    *,
    plan: _plans.PlanSimulate | None = None,
    anticipate: bool = True,
    deviation: bool = False,
) -> _np.ndarray:
    """
    """
    columns_to_run = dataslate.base_columns
    boolex_logly = dataslate.boolex_logly
    working_data = _np.copy(dataslate.data)

    column_start = columns_to_run[0]
    column_slice = slice(column_start, column_start+len(columns_to_run))
    column_array = _np.array(range(column_start, column_start+len(columns_to_run)))

    vec = solution_vectors

    T = solution.T
    R = solution.R
    K = solution.K if not deviation else 0

    Z = solution.Z
    H = solution.H
    D = solution.D if not deviation else 0

    if any(boolex_logly):
        working_data[boolex_logly, :] = _np.log(working_data[boolex_logly, :])

    curr_state = _dataslates.HorizontalDataslate.retrieve_vector_from_data_array(
        working_data, vec.transition_variables, column_start-1,
    )

    missing_initials = _np.isnan(curr_state)
    unnecessary_initials = ~_np.array(vec.initial_conditions).reshape(-1, 1)
    missing_unnecessary_initials = missing_initials & unnecessary_initials
    curr_state[missing_unnecessary_initials] = 0

    no_shift_state_to_slab_lhs = [
        t.qid
        for t in vec.transition_variables
        if t.shift == 0
    ]

    no_shift_state_to_slab_rhs = [
        j 
        for j, t in enumerate(vec.transition_variables)
        if t.shift == 0
    ]

    measurement_variables_to_slab = [t.qid for t in vec.measurement_variables]

    transition_shocks_in_slab = [t.qid for t in vec.transition_shocks]
    transition_shocks = working_data[transition_shocks_in_slab, :]

    measurement_shocks_in_slab = [t.qid for t in vec.measurement_shocks]
    measurement_shocks = working_data[[t.qid for t in vec.measurement_shocks], :]

    transition_shocks[_np.isnan(transition_shocks)] = 0
    measurement_shocks[_np.isnan(measurement_shocks)] = 0

    Rx = [R]
    forward = None

    shock_column_incidence = list(_np.any(transition_shocks[:, column_array] != 0, axis=0))
    if any(shock_column_incidence):
        # Last shock at t+forward
        forward = len(shock_column_incidence) - shock_column_incidence[::-1].index(True) - 1
        if anticipate:
            Rk = solution.expand_square_solution(forward)
        else:
            Rk = [None]*forward
        Rx = Rx + (Rk if Rk is not None else [])

    shock_column_end = (
        column_start + forward
        if forward is not None
        else column_start - 1
    )

    for t in column_array:
        shock_impact = sum( 
            Rx[k] @ transition_shocks[:, (s,)] if Rx[k] is not None else 0 
            for k, s in enumerate(range(t, shock_column_end+1))
        )
        curr_state = T @ curr_state + shock_impact + K
        working_data[no_shift_state_to_slab_lhs, t] = curr_state[no_shift_state_to_slab_rhs].flat

        y = Z @ curr_state + H @ measurement_shocks[:, (t,)] + D
        working_data[measurement_variables_to_slab, t] = y.flat

    working_data[transition_shocks_in_slab, :] = transition_shocks
    working_data[measurement_shocks_in_slab, :] = measurement_shocks

    if any(boolex_logly):
        working_data[boolex_logly, :] = _np.exp(working_data[boolex_logly, :])

    dataslate.data = working_data





"""
Interface to X13-ARIMA-TRAMO-SEATS
"""


#[
from __future__ import annotations
from typing import (Any, )
from types import (EllipsisType, )
import os as _os
import numpy as _np
import subprocess as _sp
import platform as _pf
import tempfile as _tf
import glob as _gl

from .. import executables as _executables
from .. import wrongdoings as _wrongdoings
from .. import dates as _dates
from . import main as _series
from . import _functionalize
#]


__all__ = ()


class Mixin:
    """
    """
    #[
    def x13(
        self,
        *,
        range: _dates.Ranger | EllipsisType = ...,
        when_error: _wrongdoings.HOW = "warning",
        clean_up: bool = True,
        output: str = "sa",
        mode: Literal["mult", "add", "pseudoadd", "logadd", ] = "mult",
        **kwargs,
    ) -> dict[str, Any]:
        """
        """
        range = tuple(self._resolve_dates(range))
        base_start_date = range[0]
        settings = _create_settings(base_start_date, output, mode, **kwargs, )
        spc = _create_spc_file(_TEMPLATE_SPC, settings, )
        new_data = []
        info = []
        for variant_data in self.data.T:
            ith_new_data, ith_info = _x13_data(spc, base_start_date, variant_data, settings["x11_save"], clean_up=clean_up, )
            new_data.append(ith_new_data.reshape(-1, 1, ))
            info.append(ith_info)
        if not all(i["success"] for i in info):
            _wrongdoings.raise_as(
                when_error,
                "X13 failed to produce a result for at least one variant.",
            )
        new_data = _np.hstack(new_data, )
        self._replace_start_date_and_values(base_start_date, new_data, )
        return info
    #]


for n in ("x13", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


def _x13_data(
    spc: str,
    base_start_date: _dates.Date,
    base_variant_data: _np.ndarray,
    x11_save: str,
    /,
    clean_up: bool = True,
) -> tuple[_np.ndarray, dict[str, Any]]:
    """
    """
    #[
    new_data = _np.full(base_variant_data.shape, _np.nan, dtype=float, )
    start_date, variant_data, output_slice = _remove_leading_trailing_nans(base_start_date, base_variant_data, )
    year, period = start_date.to_year_period()
    spc = spc.replace("$(series_start)", _X13_DATE_FORMAT_RESOLUTION[start_date.frequency].format(year=year, period=period, ), )
    spc = spc.replace("$(series_data)", _print_series_data(variant_data, ), )
    spc_file_name_without_ext = _write_spc_to_file(spc, )
    system_output = _execute(spc_file_name_without_ext, )
    raw_output_data, info = _collect_outputs(spc_file_name_without_ext, system_output, x11_save, )
    if raw_output_data is not None:
        new_data[output_slice] = raw_output_data
    if clean_up:
        _clean_up(spc_file_name_without_ext, )
    return new_data, info
    #]


def _print_series_data(
    variant_data: _np.ndarray,
    /,
) -> str:
    """
    """
    return "\n".join(f"        {v:g}" for v in variant_data)


def _execute(
    spc_file_name_without_ext: str,
    /,
) -> _sp.CompletedProcess:
    """
    """
    #[
    return _sp.run(
        [_X13_EXECUTABLE_PATH, spc_file_name_without_ext, ],
        stdout=_sp.PIPE,
        check=True,
    )
    #]


def _collect_outputs(
    spc_file_name_without_ext: str,
    output: _sp.CompletedProcess,
    x11_save: str,
    /,
) -> tuple[_np.ndarray | None, dict[str, str]]:
    """
    """
    #[
    info = _read_output_info(spc_file_name_without_ext, )
    info["success"] = output.returncode == 0 and "ERROR" not in str(output.stdout)
    if info["success"]:
        output_data, info[x11_save] = _read_output_data(spc_file_name_without_ext, x11_save, info, )
    else:
        output_data = None
        info[x11_save] = None
    return output_data, info
    #]


def _read_output_data(
    spc_file_name_without_ext: str,
    x11_save: str,
    info: dict[str, str],
    /,
) -> tuple[tuple[float, ...], str]:
    """
    """
    #[
    with open(spc_file_name_without_ext + "." + x11_save, "rt", ) as fid:
        table = str(fid.read())
    output_data = tuple(float(i) for i in table.split()[5::2])
    return output_data, table
    #]


def _read_output_info(
    spc_file_name_without_ext: str,
    /,
) ->  dict[str, str]:
    """
    """
    info = dict()
    for ext in ["log", "out", "err", ]:
        with open(spc_file_name_without_ext + "." + ext, "rt", ) as fid:
            info[ext] = fid.read()
    return info


def _write_spc_to_file(
    spc: str,
    /,
) -> str:
    """
    """
    with _tf.NamedTemporaryFile(dir=".", mode="wt", suffix=".spc", delete=False, ) as fid:
        spc_file_name = fid.name
        fid.write(spc)
    return spc_file_name.removesuffix(".spc")


def _remove_leading_trailing_nans(
    base_start_date: _dates.Date,
    variant_data: _np.ndarray,
    /,
) -> tuple[_dates.Date, _np.ndarray, slice]:
    """
    """
    #[
    start_date = base_start_date.copy()
    num_rows = variant_data.shape[0]
    num_leading = 0
    num_trailing = 0
    while _np.isnan(variant_data[0]):
        num_leading += 1
        variant_data = variant_data[1:]
    while _np.isnan(variant_data[-1]):
        num_trailling += 1
        variant_data = variant_data[:-1]
    return start_date + num_leading, variant_data, slice(num_leading, num_rows-num_trailing, )
    #]


def _create_settings(
    base_start_date: _dates.Date,
    output: str,
    mode: Literal["mult", "add", "pseudoadd", "logadd", ],
    /,
    transform_function: str = "none",
) ->  dict[str, str]:
    """
    """
    #[
    if str(mode) == "mult" and str(transform_function) == "none":
        transform_function = "log"
    settings = {
        "series_period": str(base_start_date.frequency.value),
        "x11_mode": str(mode),
        "x11_save": _X11_OUTPUT_RESOLUTION.get(output, output),
        "transform_function": str(transform_function),
    }
    return settings
    #]


def _create_spc_file(
    spc: str,
    settings: dict[str, str],
    /,
) -> str:
    """
    Replace placeholders with custom setting in the spc file
    """
    #[
    for k, v in settings.items():
        spc = spc.replace(f"$({k})", v)
    return spc
    #]


def _clean_up(
    spc_file_name_without_ext: str,
    /,
) -> None:
    """
    """
    #[
    for f in _gl.glob(spc_file_name_without_ext + ".*"):
        _os.remove(f)
    #]


_EXECUTABLES_PATH = _os.path.dirname(_executables.__file__)
_TEMPLATE_SPC_PATH = _os.path.join(_EXECUTABLES_PATH, "template.spc", )


with open(_TEMPLATE_SPC_PATH, "rt", ) as fid:
    _TEMPLATE_SPC = fid.read()


_X11_OUTPUT_RESOLUTION = {
    "sf": "d10",
    "sa": "d11",
    "tc": "d12",
    "irr": "d13",
}


_X13_DATE_FORMAT_RESOLUTION = {
    _dates.Frequency.MONTHLY: "{year:04d}.{period:02d}",
    _dates.Frequency.QUARTERLY: "{year:04d}.{period:1d}",
}


_X13_EXECUTABLE_FILE = {
    "Windows": "x13as.exe",
    "Linux": "x13asunix",
    "Darwin": "x13asmac",
}[_pf.system()]


_X13_EXECUTABLE_PATH = _os.path.join(_EXECUTABLES_PATH, _X13_EXECUTABLE_FILE, )





"""
Time series chartpacks
"""


#[
from __future__ import annotations

from typing import (Self, Iterator, Iterable, Sequence, Callable, )
from types import (EllipsisType, )
import re as _re
import math as _ma
import plotly.graph_objects as _pg

from .. import plotly as _plotly
from .. import dates as _dates
from ..databoxes import main as _databoxes
#]


__all__ =  (
    "Chartpack",
)


class Chartpack:
    """
    """
    #[

    def __init__(self, title: str = "", /, **kwargs, ) -> None:
        """
        """
        self.title = title
        self.span = kwargs.get("span", ..., )
        self.tiles = kwargs.get("tiles", None, )
        self.transforms = kwargs.get("transforms", None, )
        self.highlight = kwargs.get("highlight", None, )
        self.legend = kwargs.get("legend", None, )
        self._figures = []

    def plot(
        self,
        input_db: _databoxes.Databox,
        /,
        transforms: dict[str, Callable] | None = None,
        show_figure: bool = True,
    ) -> tuple[_pg.Figure, ...]:
        """
        """
        transforms = transforms if transforms is not None else self.transforms
        return tuple(
            figure.plot(input_db, **self.__dict__, )
            for figure in self
        )

    @property
    def num_figures(self, ) -> int:
        """
        """
        return len(self._figures, )

    def _add_figure(self, figure: _Figure, ) -> None:
        """
        """
        self._figures.append(figure, )

    def _add_figure_from_string(self, *args, **kwargs, ) -> None:
        """
        """
        self._add_figure(_Figure.from_string(*args, **kwargs, ), )

    def add_figure(self, figure_strings: Iterable[str] | str, ) -> None:
        """
        """
        _add_strings(self, figure_strings, self._add_figure_from_string, )

    add_figures = add_figure
    __add__ = add_figure

    def add_chart(self, *args, **kwargs, ) -> None:
        """
        """
        self._figures[-1].add_chart(*args, **kwargs, )

    add_charts = add_chart
    __lshift__ = add_chart

    def __str__(self, /, ) -> str:
        """
        """
        return repr(self, )

    def __repr__(self, /, ) -> str:
        """
        """
        return self._repr()

    def _repr(self, /, indent: str = "", ) -> str:
        """
        """
        child_indent = indent + "   |"
        lines = [f"{indent}Chartpack({self.title!r}, )", ]
        lines.extend(f"{c._repr(child_indent, )}" for c in self)
        return "\n".join(lines)

    def __getitem__(self, item: str | int, ) -> _Figure:
        """
        """
        if isinstance(item, int, ):
            return self._figures[item]
        if isinstance(item, str, ):
            return next(f for f in self._figures if f.title == item)

    def __iter__(self, ) -> Iterator[_Figure]:
        """
        """
        return iter(self._figures, )

    #]


class _Figure:
    """
    """
    #[

    __slots__ = (
        "_charts",
        "title",
    )

    def __init__(
        self,
        /,
        title: str | None = None,
    ) -> None:
        """
        """
        self._charts = []
        self.title = title

    @classmethod
    def from_string(
        klass,
        input_string: str,
        /,
    ) -> Self:
        """
        """
        return klass(title=input_string, )

    @property
    def num_charts(self, ) -> int:
        """
        """
        return len(self._charts, )

    def plot(
        self,
        input_db: _databoxes.Databox,
        /,
        span: Iterable[_dates.Dater] | EllipsisType = ...,
        tiles: Sequence[int] | int | None = None,
        transforms: dict[str, Callable] | None = None,
        show_figure: bool = True,
        highlight: tuple[_dates.Dater, ...] | None = None,
        legend: Iterable[str, ...] | None = None,
        **kwargs,
    ) -> None:
        """
        """
        tiles = _resolve_tiles(tiles, self.num_charts, )
        figure = _plotly.make_subplots(
            rows=tiles[0], columns=tiles[1],
            subplot_titles=tuple(c.caption for c in self),
        )
        for i, chart in enumerate(self, ):
            chart.plot(
                input_db,
                figure,
                i,
                span=span,
                transforms=transforms,
                legend=legend if i==0 else None,
            )
            if highlight is not None:
                _plotly.highlight(figure, highlight, subplot=i, )
        figure.update_layout(title={"text": self.title, }, )
        if show_figure:
            figure.show()
        return figure

    def __str__(self, /, ) -> str:
        """
        """
        return repr(self, )

    def __repr__(self, /, ) -> str:
        """
        """
        return self._repr()

    def _repr(self, /, indent: str = "", ) -> str:
        """
        """
        child_indent = indent + "   |"
        lines = [f"{indent}_Figure({self.title!r}, )", ]
        lines.extend(f"{c._repr(child_indent, )}" for c in self)
        return "\n".join(lines)

    def _add_chart(self, chart: _Chart, ) -> None:
        """
        """
        self._charts.append(chart, )

    def _add_chart_from_string(self, input_string: str, /, ) -> None:
        """
        """
        self._add_chart(_Chart.from_string(input_string, ), )

    def add_chart(self, chart_strings: Iterable[str] | str, ) -> None:
        """
        """
        _add_strings(self, chart_strings, self._add_chart_from_string, )

    add_charts = add_chart
    __lshift__ = add_chart

    def __iter__(self, /, ) -> Iterator[_Chart]:
        """
        """
        return iter(self._charts, )

    #]


class _Chart:
    """
    """
    #[

    _INPUT_STRING_PATTERN = _re.compile(r"^(?:(?P<title>.*?):)?(?P<expression>.*?)(?:\[(?P<transform>.*?)\])?$")

    __slots__ = (
        "title",
        "_expression",
        "_transform"
    )

    def __init__(
        self,
        expression: str,
        title: str | None = None,
        transform: str | Callable | None = None,
    ) -> None:
        """
        """
        self._expression = (expression or "").strip()
        self.title = (title or "").strip()
        self._transform = (transform or "").strip()

    @classmethod
    def from_string(
        klass: type(Self),
        input_string: str,
        /,
    ) -> Self:
        """
        """
        match = klass._INPUT_STRING_PATTERN.match(input_string, )
        if not match:
            raise ValueError(f"Invalid input string: {input_string!r}")
        return klass(**match.groupdict(), )

    @property
    def caption(self, ) -> str:
        """
        """
        return self.title or (self._expression + f" [{self._transform}]" if self._transform else "")

    def plot(
        self,
        input_db: _databoxes.Databox,
        figure: _pg.Figure,
        index: int,
        /,
        span: Iterable[_dates.Dater] | EllipsisType = ...,
        transforms: dict[str, Callable] | None = None,
        legend: Iterable[str, ...] | None = None,
    ) -> None:
        """
        """
        series_to_plot = input_db.evaluate_expression(self._expression, )
        series_to_plot = self._apply_transform(series_to_plot, transforms, )
        series_to_plot.plot(
            figure=figure,
            subplot=index,
            span=span,
            show_figure=False,
            freeze_span=True,
            legend=legend,
        )

    def _apply_transform(self, x, transforms, ):
        """
        """
        if self._transform:
            func = transforms[self._transform]
        else:
            func = None
        return func(x) if func else x


    def __str__(self, /, ) -> str:
        """
        """
        return repr(self, )

    def __repr__(self, /, ) -> str:
        """
        """
        return self._repr()

    def _repr(self, /, indent: str = "", ) -> str:
        """
        """
        return f"{indent}_Chart({self.title!r}, {self._expression!r}, {self._transform!r}, )"

    #]


def _add_strings(self, input_strings: Iterable[str] | str, call: Callable ) -> None:
    """
    """
    #[
    if isinstance(input_strings, str, ):
        input_strings = (input_strings, )
    for s in input_strings:
        call(s, )
    #]


def _resolve_tiles(tiles, num_charts, ) -> tuple[int, int]:
    """
    """
    #[
    if tiles is None:
        return _auto_tiles(num_charts, )
    if isinstance(tiles, int, ):
        return _auto_tiles(tiles, )
    if isinstance(tiles, Sequence, ):
        return tiles[:2]
    raise TypeError(f"Invalid tiles: {tiles!r}")
    #]


def _auto_tiles(num_charts, ) -> tuple[int, int]:
    """
    """
    #[
    n = _ma.ceil(_ma.sqrt(num_charts, ), )
    if n * (n-1) >= num_charts:
        return (n, n-1, )
    return (n, n, )
    #]






import sys
sys.path.append("../../..")

from modiphy.dataman.series import Series
from modiphy.dataman.dates import yy, hh, qq, mm, dd, ii

x = Series()
x[ qq(2021,1)>>qq(2024,4) ] = [ i**2 for i in range(16) ]


y = Series(num_columns=2)
y[ qq(2020,1)>>qq(2025,4), 0 ] = 100
y[ qq(2025,1)>>qq(2027,4), 1 ] = -100

print(x)
print(y)
print(y(-1)) # Matlab y{-1}
print(x | y) # Matlab [x, y]
print(y(-1) | 3) # Matlab [y{-1}, 3]

data1 = 0.01 * y[..., 1] # Matlab 0.01 * y{:, 1}
print(data1)

data = 0.01 * y[..., ...] # Matlab 0.01 * y{:, :}
print(data)

print(y.start_date)
print(y.end_date)
print(y.range)

range = y.range
print(range.to_plotly_dates())




"""
Frequency conversion of time series
"""


#[
from __future__ import annotations

from typing import (Self, Callable, )
from types import (EllipsisType, )
import numpy as _np
import functools as _ft
import operator as _op
import statistics as _st

from .. import dates as _dates
from . import arip as _arip
from . import _functionalize
#]


__all__ = (
    "convert_roc",
    "convert_pct",
)


_builtin_min = min
_builtin_max = max
_builtin_sum = sum


class ConversionMixin:
    """
    """
    #[

    def aggregate(
        self,
        new_freq: _dates.Frequency,
        /,
        method: Literal["mean", "sum", "first", "last", "min", "max"] = "mean",
        remove_missing: bool = False,
        select: list[int] | None = None,
    ) -> None:
        """
        ---

        # `aggregate`

        Aggregate time series to a lower frequency

        ## Syntax

            y = x.aggregate(new_freq, method="mean", remove_missing=False, select=None)

        ## Input arguments

            * `new_freq`: Frequency
            >
            > The new frequency to aggregate the original time series to.
            >

            * `method` {"mean", "sum", "first", "last", "min", "max"}
            >
            > The method to use for aggregation. The following methods are available:
            >
        """
        method_func = (
            _AGGREGATION_METHOD_RESOLUTION[method]
            if isinstance(method, str) else method
        )
        #
        if new_freq == self.frequency:
            return
        if new_freq > self.frequency or new_freq is _dates.Frequency.UNKNOWN or self.frequency is _dates.Frequency.UNKNOWN:
            raise ValueError(f"Cannot aggregate from {self.frequency} frequency to {new_freq} frequency")
        #
        new_dater_class = _dates.DATER_CLASS_FROM_FREQUENCY_RESOLUTION[new_freq]
        #
        aggregate_within_data_func = _ft.partial(
            _aggregate_within_data,
            select,
            remove_missing,
            method_func,
        )
        #
        if self.frequency.is_regular:
            aggregate_func = _aggregate_regular_to_regular
        elif self.frequency is _dates.Frequency.DAILY:
            aggregate_func = _aggregate_daily_to_regular
        #
        new_start_date, new_data = aggregate_func(self, new_dater_class, aggregate_within_data_func, )
        self._replace_start_date_and_values(new_start_date, new_data, )

    def disaggregate(
        self,
        target_freq: _dates.Frequency,
        /,
        method: str = "flat",
        **kwargs,
    ) -> Self:
        """
        Disaggregate time series to a higher frequency
        """
        method_func = _CHOOSE_DISAGGREGATION_METHOD[method]
        #
        if target_freq == self.frequency:
            return
        #
        if target_freq < self.frequency \
            or target_freq is _dates.Frequency.UNKNOWN \
            or self.frequency is _dates.Frequency.UNKNOWN:
            #
            raise _wrongdoings.IrisPieCritical(
                f"Cannot disaggregate from {self.frequency} frequency to {target_freq} frequency"
            )
        #
        new_dater_class = _dates.DATER_CLASS_FROM_FREQUENCY_RESOLUTION[target_freq]
        new_start_date, new_data, *_ = method_func(self, new_dater_class, **kwargs, )
        self._replace_start_date_and_values(new_start_date, new_data, )

    #]


for n in ("aggregate", "disaggregate", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


def _aggregate_daily_to_regular(
    self,
    new_dater_class: type,
    aggregate_within_data_func: Callable,
) -> Self:
    """
    """
    #[
    start_date = self.start_date.create_soy()
    end_date = self.end_date.create_eoy()
    from_to = (start_date, end_date, )
    start_year = self.start_date.get_year()
    end_year = self.end_date.get_year()
    new_start_date = new_dater_class.from_year_period(start_year, 1)
    new_end_date = new_dater_class.from_year_period(end_year, "end")
    get_slice_func = lambda t: slice(
        t.to_daily(position="start") - start_date,
        t.to_daily(position="end") - start_date + 1,
    )
    new_data = tuple(
        tuple(
            aggregate_within_data_func(variant[get_slice_func(t)])
            for data_variant in self.iter_own_data_variants_from_to(from_to, )
        )
        for t in _dates.Ranger(new_start_date, new_end_date)
    )
    #
    new_data = _np.array(new_data, dtype=self.data_type, )
    return new_start_date, new_data
    #]


def _aggregate_regular_to_regular(
    self,
    new_dater_class: type,
    aggregate_within_data_func: Callable,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    start_year = self.start_date.get_year()
    start_date = self.start_date.create_soy()
    end_date = self.end_date.create_eoy()
    new_start_date = new_dater_class.from_year_period(start_year, 1)
    new_freq = new_dater_class.frequency
    self_data = self.get_data_from_to((start_date, end_date))
    factor = self.frequency.value // new_freq
    #
    # Loop over variants, create within-period data, and aggregate
    multi_output_data_transposed = []
    for data_variant in self_data.T:
        output_data_transposed = tuple(
            aggregate_within_data_func(within_data)
            for within_data in data_variant.reshape((-1, factor), )
        )
        multi_output_data_transposed.append(output_data_transposed, )
    #
    multi_output_data = _np.array(multi_output_data_transposed, dtype=self.data_type, ).T
    return new_start_date, multi_output_data
    #]


def _disaggregate_flat(
    self,
    high_dater_class: type,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    high_freq = high_dater_class.frequency
    high_start_date = high_dater_class.from_ymd(*self.start_date.to_ymd(position="start", ), )
    factor = high_freq.value // self.frequency.value
    high_data = _np.repeat(self.data, factor, axis=0, )
    return high_start_date, high_data, factor
    #]


def _disaggregate_first(
    self,
    high_dater_class: type,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    high_start_date, high_data, factor = _disaggregate_flat(self, high_dater_class, )
    for i in range(1, factor, ):
        high_data[i::factor, :] = _np.nan
    return high_start_date, high_data
    #]


def _disaggregate_last(
    self,
    high_dater_class: type,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    high_start_date, high_data, factor = _disaggregate_flat(self, high_dater_class, )
    for i in range(0, factor-1, ):
        high_data[i::factor, :] = _np.nan
    return high_start_date, high_data, factor
    #]


def _aggregate_within_data(
    select: list[int] | None,
    remove_missing: bool,
    method_func: Callable,
    within_data: _np.ndarray,
    /,
) -> _np.ndarray:
    """
    """
    #[
    if select is not None:
        select = tuple(select)
        within_data = within_data[select]
    if remove_missing:
        within_data = within_data[~_np.isnan(within_data)]
    return method_func(within_data) if within_data.size > 0 else _np.nan
    #]


_AGGREGATION_METHOD_RESOLUTION = {
    "mean": _st.mean,
    "geometric_mean": _st.geometric_mean,
    "sum": _builtin_sum,
    "first": _op.itemgetter(0),
    "last": _op.itemgetter(-1),
    "min": _builtin_min,
    "max": _builtin_max,
}


_CHOOSE_DISAGGREGATION_METHOD = {
    "flat": _disaggregate_flat,
    "first": _disaggregate_first,
    "last": _disaggregate_last,
    "arip": _arip.disaggregate_arip,
}


def convert_roc(
    roc: Real,
    from_freq: _dates.Frequency,
    to_freq: _dates.Frequency,
    /,
) -> Real:
    """
    """
    return float(roc) ** (float(from_freq) / float(to_freq))


def convert_pct(
    pct: Real,
    from_freq: _dates.Frequency,
    to_freq: _dates.Frequency,
    /,
) -> Real:
    """
    """
    return 100*(convert_roc(1 + float(pct)/100, from_freq, to_freq) - 1)


def convert_diff(
    diff: Real,
    from_freq: _dates.Frequency,
    to_freq: _dates.Frequency,
    /,
) -> Real:
    """
    """
    return float(diff) * (float(from_freq) / float(to_freq))




"""
"""


#[
from __future__ import annotations

from collections.abc import (Generator, Iterable, )
from typing import (Any, )
import numpy as _np
import dataclasses as _dc

from .. import wrongdoings as _wrongdoings
#]


@_dc.dataclass
class _Block:
    """
    """
    #[
    eids: tuple[int, ...] | None = None
    qids: tuple[int, ...] | None = None

    @property
    def num_equations(self, ) -> int:
        return len(self.eids)

    @property
    def num_quantities(self, ) -> int:
        return len(self.qids)
    #]


def blaze(
    im: _np.ndarray,
    /,
    eids: Iterable[int] = None,
    qids: Iterable[int] = None,
) -> tuple[tuple[_Block, ...], dict[str, Any]]:
    """
    Complete sequential block analysis of incidence matrix
    """
    #[
    if eids is None:
        eids = range(im.shape[0])
    if qids is None:
        qids = range(im.shape[1])
    eids = tuple(eids)
    qids = tuple(qids)
    #
    # Step 1: Prefetch quantities/equations that can be ordered first
    # individually (an equation only has one quantity), and ordered last
    # individually (a quantity only occurs in one equation)
    #
    eids_first, qids_first, eids_last, qids_last, eids_inner, qids_inner, im_inner \
    = prefetch(im, eids=eids, qids=qids, )
    #
    # Step 2: If there is a remaining core of interdependent equations, run
    # a naive triangularization algorithm on it
    #
    if im_inner.size > 0:
        eids_inner, qids_inner, im_inner, *_ \
        = triangularize_inner_block(im_inner, eids=eids_inner, qids=qids_inner, )
    #
    # Combine the results
    #
    eids = eids_first + eids_inner + eids_last
    qids = qids_first + qids_inner + qids_last
    #
    # Step 3: Create a tuple with a Block object for each block of
    # equations and equantities
    #
    first_blocks = tuple(_Block(eid, qid, ) for eid, qid in zip(eids_first, qids_first, ))
    inner_blocks = tuple(_generate_inner_blocks(im_inner, eids=eids_inner, qids=qids_inner, ))
    last_blocks = tuple(_Block(eid, qid, ) for eid, qid in zip(eids_last, qids_last, ))
    #
    info = {
        "eids_first": eids_first,
        "qids_first": qids_first,
        "eids_last": eids_last,
        "qids_last": qids_last,
        "eids_inner": eids_inner,
        "qids_inner": qids_inner,
        "im_inner": im_inner,
    }
    return first_blocks + inner_blocks + last_blocks, info
    #]


def _generate_inner_blocks(
    im: _np.ndarray,
    eids: Iterable[int],
    qids: Iterable[int],
) -> Generator[_Block]:
    """
    """
    #[
    eids = tuple(eids)
    qids = tuple(qids)
    while im.size:
        block_size = next(i for i in range(1, im.shape[0] + 1) if not im[:i, i:].any())
        yield _Block(eids[:block_size], qids[:block_size], )
        eids = eids[block_size:]
        qids = qids[block_size:]
        im = im[block_size:, block_size:]
    #]


def prefetch(
    im: _np.ndarray,
    /,
    eids: Iterable[int] = None,
    qids: Iterable[int] = None,
) -> tuple[
    tuple[int, ...],
    tuple[int, ...],
    tuple[int, ...],
    tuple[int, ...],
    tuple[int, ...],
    tuple[int, ...],
    _np.ndarray,
]:
    """
    Recursively prefetch singleton equations and quantities that can be ordered first and last
    """
    #[
    initial_size = im.size
    if eids is None:
        eids = tuple(range(im.shape[0]))
    if qids is None:
        qids = tuple(range(im.shape[1]))
    eids_first, qids_first, eids, qids, im = _prefetch_first(im, eids, qids, )
    eids_last, qids_last, eids, qids, im = _prefetch_last(im, eids, qids, )
    if im.size < initial_size:
        eids_first_next, qids_first_next, \
            eids_last_next, qids_last_next, \
            eids, qids, im, \
            = prefetch(im, eids=eids, qids=qids, )
        #
        eids_first = eids_first + eids_first_next
        qids_first = qids_first + qids_first_next
        #
        eids_last = eids_last_next + eids_last
        qids_last = qids_last_next + qids_last
        #
    return \
        eids_first, qids_first, \
        eids_last, qids_last, \
        eids, qids, im,
    #]


def sequentialize_strictly(
    im: _np.ndarray,
    /,
    eids: Iterable[int] | None = None,
    qids: Iterable[int] | None = None,
) -> tuple[int, ...]:
        """
        """
        #[
        if eids is None:
            eids = range(im.shape[0])
        if qids is None:
            qids = range(im.shape[1])
        eids = tuple(eids)
        qids = tuple(qids)
        #
        eids_first, qids_first, \
            eids_last, qids_last, \
            eids_rem, qids_rem, im_rem, \
            = prefetch(im, eids=eids, qids=qids, )
        #
        fail = \
            im_rem.size or eids_rem or qids_rem \
            or eids_first != qids_first \
            or eids_last != qids_last
        if fail:
            _wrongdoings.IrisPieError("Cannot find strict sequential reordering", )
        return eids_first + eids_last
        #]


def _prefetch_first(im, eids, qids, /, ):
    """
    Equations with a single incidence of quantities are reordered first
    """
    sum_in_rows = im.sum(axis=1, )
    index_rows = tuple(_np.where(sum_in_rows == 1)[0])
    index_columns = [
        _np.where(im[i, :] == 1)[0][0]
        for i in index_rows
    ]
    im = _np.delete(im, index_rows, axis=0, )
    im = _np.delete(im, index_columns, axis=1, )
    eids_first, eids_rem = _split_ids(eids, index_rows, )
    qids_first, qids_rem = _split_ids(qids, index_columns, )
    return eids_first, qids_first, eids_rem, qids_rem, im


def _prefetch_last(im, eids, qids, /, ):
    """
    Quantities with a single incidence in equations are reordered last
    """
    sum_in_columns = im.sum(axis=0, )
    index_columns = tuple(_np.where(sum_in_columns == 1)[0])
    index_rows = [
        _np.where(im[:, j] == 1)[0][0]
        for j in index_columns
    ]
    im = _np.delete(im, index_columns, axis=1, )
    im = _np.delete(im, index_rows, axis=0, )
    eids_last, eids_rem = _split_ids(eids, index_rows, )
    qids_last, qids_rem = _split_ids(qids, index_columns, )
    return eids_last, qids_last, eids_rem, qids_rem, im


def _split_ids(ids, index):
    extracted = tuple(ids[i] for i in index)
    remaining = tuple(id_ for i, id_ in enumerate(ids) if i not in index)
    return extracted, remaining


def triangularize_inner_block(
    im: _np.ndarray,
    /,
    eids: Iterable[int] = None,
    qids: Iterable[int] = None,
    max_iterations: int = 100,
) -> tuple[tuple[int, ...], tuple[int, ...], _np.ndarray, dict[str, Any]]:
    """
    Naive block triangularization
    """
    #[
    im = im.copy()
    if eids is None:
        eids = tuple(_np.arange(im.shape[0]))
    if qids is None:
        qids = tuple(_np.arange(im.shape[1]))
    #
    eids = _np.array(eids, dtype=int, )
    qids = _np.array(qids, dtype=int, )
    count = 0
    while True:
        #
        # Check if too many iterations
        if count >= max_iterations:
            break
        #
        count += 1
        prev_eids = eids
        prev_qids = qids
        #
        # Reorder columns (quantities) given rows (equations)
        column_reordering = _get_column_reordering(im)
        im = im[:, column_reordering]
        qids = qids[column_reordering]
        #
        # Reorder rows (equations) given columns (quantities)
        row_reordering = _get_row_reordering(im)
        im = im[row_reordering, :]
        eids = eids[row_reordering]
        #
        # Check if no change
        if _np.array_equal(prev_eids, eids) and _np.array_equal(prev_qids, qids):
            break
        #
    eids_reordered = tuple(eids)
    qids_reordered = tuple(qids)
    info = {"iterations": count}
    return eids_reordered, qids_reordered, im, info
    #]


def _get_column_reordering(
    im: _np.ndarray,
    /,
) -> _np.ndarray:
    """
    """
    #[
    num_incidences_in_columns = _np.sum(im, axis=0, )
    return _np.flip(_np.argsort(num_incidences_in_columns))
    #]


def _get_row_reordering(
    im: _np.ndarray,
    /,
) -> _np.ndarray:
    """
    """
    #[
    num_incidences_in_rows = _np.sum(im, axis=1, )
    return _np.argsort(num_incidences_in_rows)
    #]


def is_sequential(
    im: _np.ndarray,
    /,
    order: int = 1,
) -> bool:
    """
    """
    return _np.all(~_np.triu(im, order))




"""
"""


#[
from __future__ import annotations

from typing import (Any, )
from numbers import (Real, )
import numpy as _np

from .. import wrongdoings as _wrongdoings
from ..databoxes import main as _databoxes
from ..plans import main as _plans
from ..explanatories import main as _explanatories
from .. import dataslates as _dataslates
#]


_DATASLATE_VARIANT_ITERATOR = \
    _dataslates.HorizontalDataslate.iter_variants_from_databox_for_slatable


class SimulateMixin:
    """
    """
    #[

    def simulate(
        self,
        in_databox: _databoxes.Databox,
        span: Iterable[Dater],
        /,
        plan: _plans.Plan | None = None,
        prepend_input: bool = True,
        target_databox: _databoxes.Databox | None = None,
        when_nonfinite: Literal["error", "warning", "silent", ] = "warning",
    ) -> tuple[_databoxes.Databox, dict[str, Any]]:
        """
        """
        #
        # Check consistency of the plan and the model/simulation
        # Add extra custom databox names to the model
        base_dates = tuple(span, )
        if plan is not None:
            plan.check_consistency(self, base_dates, )
            self.set_extra_databox_names(plan.get_databox_names(), )
        #
        out_dataslates = []
        num_variants = 1
        zipped = zip(
            range(num_variants, ),
            self.iter_variants(),
            _DATASLATE_VARIANT_ITERATOR(self, in_databox, base_dates, ),
        )
        #
        #=======================================================================
        # Main loop over variants
        for vid, mdi, dsi in zipped:
            mdi._simulate_periods_and_explanatories(
                dsi,
                plan=plan,
                when_nonfinite=when_nonfinite,
            )
            dsi.remove_terminal()
            out_dataslates.append(dsi, )
        #=======================================================================
        #
        # Build output databox
        self.set_extra_databox_names(None, )
        out_db = _dataslates.multiple_to_databox(out_dataslates, )
        #
        if prepend_input:
            out_db.prepend(in_databox, base_dates[0]-1, )
        #
        # Add to custom databox
        if target_databox is not None:
            out_db = target_databox | out_db
        #
        info = {}
        #
        return out_db, info

    def _simulate_periods_and_explanatories(
        self,
        ds: _dataslates.HorizontalDataslate,
        /,
        plan: _plans.Plan | None = None,
        when_nonfinite: Literal["error", "warning", "silent", ] = "warning",
    ) -> None:
        """
        """
        when_nonfinite_stream = \
            _wrongdoings.STREAM_FACTORY[when_nonfinite] \
            ("Simulating the following data point(s) resulted in non-finite values:", )
        #
        name_to_row = ds.create_name_to_row()
        first_base_column = (
            ds.base_columns[0]
            if ds.base_columns
            else None
        )
        #
        for data_column in ds.base_columns:
            plan_column = data_column - first_base_column
            date = ds.column_dates[data_column]
            #
            for x in self.iter_explanatories():
                #
                transform = (
                    _get_transform(plan, x.lhs_name, plan_column, )
                    if not x.is_identity
                    else None
                )
                #
                is_exogenized, implied_value = _is_exogenized(x.lhs_name, transform, ds.data, data_column, name_to_row, )
                if is_exogenized:
                    info = x.exogenize(ds.data, data_column, implied_value)
                else:
                    info = x.simulate(ds.data, data_column, )
                _catch_nonfinite(when_nonfinite_stream, info["is_finite"], x.lhs_name, date, )
                #
        when_nonfinite_stream._raise()

    #]


def _get_transform(
    plan: _plans.Plan | None,
    lhs_name: str,
    plan_column: int,
) -> _plans.Transform | None:
    """
    """
    return (
        plan.get_exogenized_point(lhs_name, plan_column, )
        if plan is not None else None
    )


def _is_exogenized(
    lhs_name: str,
    transform: _plans.Transform | None,
    data: _np.ndarray,
    data_column: int,
    name_to_row: dict[str, int],
) -> tuple[bool, Real | None]:
    """
    """
    #[
    if transform is None:
        return False, None
    #
    lhs_name_row = name_to_row[lhs_name]
    values_before = data[lhs_name_row, :data_column]
    #
    transform_name = transform.resolve_databox_name(lhs_name, )
    transform_row = name_to_row.get(transform_name, None, )
    transform_values_after = data[transform_row, data_column:] if transform_row is not None else None
    implied_value = transform.eval_exogenized(transform_values_after, values_before, )
    if transform.when_data and _np.isnan(implied_value):
        return False, None
    return True, implied_value
    #]


def _catch_nonfinite(
    stream: _wrongdoings.Stream,
    is_finite: bool | _np.ndarray,
    name: str,
    date: Dater,
) -> None:
    """
    """
    #[
    if _np.all(is_finite, ):
        return
    message = f"{name}[{date}]"
    stream.add(message, )
    #]




"""
Description mixin
"""


#[
from __future__ import annotations
#]


class DescriptionMixin:
    """
    """
    #[

    def get_description(self, /, ) -> str:
        return str(self._description or "")

    def set_description(self, description: str, /, ) -> None:
        self._description = str(description or "")

    #]




"""
Simultaneous models
"""


#[
from __future__ import annotations

from typing import (Self, Any, TypeAlias, Literal, )
from types import EllipsisType
from collections.abc import (Iterable, Iterator, Callable, )
from numbers import (Number, )
import copy as _co
import numpy as _np
import itertools as _it
import functools as _ft

from .. import equations as _equations
from .. import quantities as _quantities
from .. import sources as _sources
from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from .. import has_invariant as _has_invariant
from .. import has_variants as _has_variants
from ..conveniences import iterators as _iterators
from ..conveniences import files as _files
from ..parsers import common as _pc
from ..databoxes import main as _databoxes
from ..fords import solutions as _solutions
from ..fords import steadiers as _fs
from ..fords import descriptors as _descriptors
from ..fords import systems as _systems

from . import invariants as _invariants
from . import variants as _variants
from . import _covariances as _covariances
from . import _flags as _flags
from . import _simulate as _simulate
from . import _steady as _steady
from . import _get as _get
from . import _assigns as _assigns
#]


__all__ = [
    "Simultaneous",
    "Model",
]


_SIMULATE_CAN_BE_EXOGENIZED = _quantities.QuantityKind.ENDOGENOUS_VARIABLE
_SIMULATE_CAN_BE_ENDOGENIZED = _quantities.QuantityKind.EXOGENOUS_VARIABLE | _quantities.QuantityKind.SHOCK
_SIMULATE_CAN_BE_ANTICIPATED = _SIMULATE_CAN_BE_EXOGENIZED | _SIMULATE_CAN_BE_ENDOGENIZED


class Simultaneous(
    _sources.SourceMixin,
    _assigns.AssignMixin,
    _has_invariant.HasInvariantMixin,
    _has_variants.HasVariantsMixin,
    _simulate.SimulateMixin,
    _steady.SteadyMixin,
    _covariances.CoverianceMixin,
    _get.GetMixin,
    _files.FromFileMixin,
):
    """

    """
    #[

    __slots__ = (
        "_invariant",
        "_variants",
    )

    def __init__(
        self,
        /,
        skeleton: bool = False,
    ) -> None:
        """
        """
        self._invariant = None
        self._variants = []

    @classmethod
    def skeleton(
        klass,
        /,
    ) -> Self:
        """
        """
        return klass(skeleton=True, )

    def copy(self, /, ) -> Self:
        """
        Create a deep copy of this model
        """
        return _co.deepcopy(self)

    def __getitem__(
        self,
        request,
        /,
    ):
        """
        Implement self[i] for variants and self[name] for quantitie values
        """
        if isinstance(request, str):
            return self.get_value(request, )
        else:
            return self.get_variant(request, )

    def __setitem__(
        self,
        reference,
        value: Any,
        /,
    ) -> None:
        """
        """
        if isinstance(reference, str):
            self.assign(**{reference: value}, )

    def __repr__(self, /, ) -> str:
        """
        """
        indented = " " * 4
        min_shift, max_shift = self.get_min_max_shifts()
        return "\n".join((
            f"",
            f"{self.__class__.__name__} model",
            f"Description: \"{self.get_description()}\"",
            f"|",
            f"| Num of variants: {self.num_variants}",
            f"| Num of equations [transition, measurement]: [{self.num_transition_equations}, {self.num_measurement_equations}]",
            f"| Time shifts [min, max]: [{min_shift:+g}, {max_shift:+g}]",
            f"|",
        ))

    def get_value(
        self,
        name: str,
        /,
    ) -> Any:
        """
        """
        names = (name, )
        quantities = self._invariant.quantities
        qids, invalid_names = _quantities.lookup_qids_by_name(quantities, names, )
        if invalid_names:
            raise _wrongdoings.IrisPieError(f"Invalid model name \"{invalid_names[0]}\"", )
        else:
            singleton_dict = self._get_values("levels", qids, )
            return singleton_dict[name]

    def change_logly(
        self,
        new_logly: bool,
        some_names: Iterable[str] | None = None,
        /
    ) -> None:
        """
        Change the log-status of some model quantities
        """
        some_names = set(some_names) if some_names else None
        qids = [
            qty.id
            for qty in self._invariant.quantities
            if qty.logly is not None and (some_names is None or qty.human in some_names)
        ]
        self._invariant.quantities = _quantities.change_logly(self._invariant.quantities, new_logly, qids)

    @property
    def is_linear(self, /, ) -> bool:
        """
        True for models declared as linear
        """
        return self._invariant._flags.is_linear

    @property
    def is_flat(self, /, ) -> bool:
        """
        True for models declared as flat
        """
        return self._invariant._flags.is_flat

    @property
    def num_transition_equations(self, /, ) -> int:
        """
        """
        return self._invariant.num_transition_equations

    @property
    def num_measurement_equations(self, /, ) -> int:
        """
        """
        return self._invariant.num_measurement_equations

    def create_name_to_qid(self, /, ) -> dict[str, int]:
        return _quantities.create_name_to_qid(self._invariant.quantities)

    def create_qid_to_name(self, /, ) -> dict[int, str]:
        return _quantities.create_qid_to_name(self._invariant.quantities)

    def create_qid_to_kind(self, /, ) -> dict[int, str]:
        return _quantities.create_qid_to_kind(self._invariant.quantities)

    def create_qid_to_description(self, /, ) -> dict[int, str]:
        """
        Create a dictionary mapping from quantity id to quantity descriptor
        """
        return _quantities.create_qid_to_description(self._invariant.quantities)

    def create_qid_to_logly(self, /, ) -> dict[int, bool]:
        """
        Create a dictionary mapping from quantity id to quantity log-status
        """
        return _quantities.create_qid_to_logly(self._invariant.quantities)

    def create_steady_array(
        self,
        /,
        variant: _variants.Variant | None = None,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        qid_to_logly = self.create_qid_to_logly()
        if variant is None:
            variant = self._variants[0]
        return variant.create_steady_array(qid_to_logly, **kwargs, )

    def create_zero_array(
        self,
        /,
        variant: _variants.Variant | None = None,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        qid_to_logly = self.create_qid_to_logly()
        if variant is None:
            variant = self._variants[0]
        return variant.create_zero_array(qid_to_logly, **kwargs, )

    def create_some_array(
        self,
        /,
        deviation: bool,
        **kwargs,
    ) -> _np.ndarray:
        return {
            True: self.create_zero_array, False: self.create_steady_array,
        }[deviation](**kwargs)

    def _enforce_assignment_rules(self, variant, /, ) -> None:
        """
        """
        #
        # Reset levels of shocks to zero
        #
        name_to_qid = self.create_name_to_qid()
        shock_qids = _quantities.generate_qids_by_kind(
            self._invariant.quantities, _quantities.QuantityKind.SHOCK,
        )
        zero_shocks = { i: 0 for i in shock_qids }
        variant.update_values_from_dict(zero_shocks, )
        #
        # Remove changes from quantities that are not logly variables
        #
        nonloglies = _quantities.generate_qids_by_kind(
            self._invariant.quantities, ~_sources.LOGLY_VARIABLE,
        )
        assign_nonloglies = { qid: (..., _np.nan, ) for qid in nonloglies }
        variant.update_values_from_dict(assign_nonloglies, )

    def systemize(
        self,
        /,
        **kwargs,
    ) -> Iterable[_systems.System]:
        """
        Create unsolved first-order system for each variant
        """
        model_flags = self._invariant._flags.update_from_kwargs(**kwargs, )
        return tuple(
            self._systemize(variant, self._invariant.dynamic_descriptor, model_flags, )
            for variant in self._variants
        )

    def _systemize(
        self,
        variant: _variants.Variant,
        descriptor: _descriptors.Descriptor,
        model_flags: flags.Flags,
        /,
    ) -> _systems.System:
        """
        Create unsolved first-order system for one variant
        """
        min_shift, max_shift = self.get_min_max_shifts()
        num_columns = -min_shift + 1 + max_shift
        qid_to_logly = self.create_qid_to_logly()
        #
        if model_flags.is_linear:
            data_array = variant.create_zero_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=min_shift, )
            data_array_lagged = None
            steady_array = variant.create_steady_array(qid_to_logly, num_columns=1, ).reshape(-1)
        else:
            data_array = variant.create_steady_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=min_shift, )
            data_array_lagged = variant.create_steady_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=min_shift-1, )
            steady_array = data_array[:, -min_shift]
        #
        column_offset = -min_shift
        return _systems.System(
            descriptor, data_array, steady_array,
            model_flags, data_array_lagged,
            column_offset,
        )

    def solve(
        self,
        /,
        **kwargs,
    ) -> None:
        """
        Calculate first-order solution for each variant within this model
        """
        model_flags = self._invariant._flags.update_from_kwargs(**kwargs, )
        for variant in self._variants:
            self._solve(variant, model_flags, )

    def _solve(
        self,
        variant: _variants.Variant,
        model_flags: flags.Flags,
        /,
    ) -> None:
        """
        Calculate first-order solution for one variant of this model
        """
        system = self._systemize(variant, self._invariant.dynamic_descriptor, model_flags, )
        variant.solution = _solutions.Solution(self._invariant.dynamic_descriptor, system, )

    def _choose_plain_equator(
        self,
        equation_switch: Literal["dynamic", "steady", ],
        /,
    ) -> Callable | None:
        """
        """
        match equation_switch:
            case "dynamic":
                return self._invariant._plain_dynamic_equator
            case "steady":
                return self._invariant._plain_steady_equator

    def _assign_default_stds(self, /, ) -> None:
        """
        Initialize standard deviations of shocks to default values
        """
        std_names = _quantities.generate_quantity_names_by_kind(self._invariant.quantities, _quantities.QuantityKind.STD, )
        dict_to_assign = { k: self._invariant._default_std for k in std_names }
        self.assign(**dict_to_assign, )

    @classmethod
    def from_source(
        klass,
        source: _sources.ModelSource,
        /,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass()
        self._invariant = _invariants.Invariant(source, **kwargs, )
        initial_variant = _variants.Variant(
            self._invariant.quantities,
            self._invariant._flags.is_flat,
        )
        self._variants = [ initial_variant ]
        self._assign_default_stds()
        self._enforce_assignment_rules(self._variants[0], )
        return self

    def get_context(self, /, ) -> dict[str, Any]:
        """
        """
        return self._invariant._context

    #
    # ===== Implement SlatableProtocol =====
    #

    def get_min_max_shifts(self, ) -> tuple[int, int]:
        """
        """
        return self._invariant._min_shift, self._invariant._max_shift

    def get_databox_names(self, /, ) -> tuple[str, ...]:
        """
        """
        qid_to_name = self.create_qid_to_name()
        return tuple(qid_to_name[qid] for qid in sorted(qid_to_name))

    def get_fallbacks(self, /, ) -> dict[str, Number]:
        """
        """
        shocks = _quantities.generate_quantity_names_by_kind(self._invariant.quantities, _quantities.QuantityKind.SHOCK)
        return { name: 0 for name in shocks }

    def get_overwrites(self, /, ) -> dict[str, Any]:
        """
        """
        return self.get_parameters_stds()

    def get_scalar_names(self, /, ) -> tuple[str, ...]:
        """
        """
        return _quantities.generate_quantity_names_by_kind(
            self._invariant.quantities,
            _quantities.QuantityKind.PARAMETER | _quantities.QuantityKind.STD,
        )

    #
    # ===== Implement PlannableSimulateProtocol =====
    #

    @property
    def simulate_can_be_exogenized(self, /, ) -> tuple[str, ...]:
        return tuple(_quantities.generate_quantity_names_by_kind(
            self._invariant.quantities, _SIMULATE_CAN_BE_EXOGENIZED,
        ))

    @property
    def simulate_can_be_endogenized(self, /, ) -> tuple[str, ...]:
        return tuple(_quantities.generate_quantity_names_by_kind(
            self._invariant.quantities, _SIMULATE_CAN_BE_ENDOGENIZED,
        ))

    @property
    def simulate_can_be_anticipated(self, /, ) -> tuple[str, ...]:
        return tuple(_quantities.generate_quantity_names_by_kind(
            self._invariant.quantities, _SIMULATE_CAN_BE_ANTICIPATED,
        ))

    simulate_can_be_when_data = ()

    #]


Model = Simultaneous




"""
"""

#[
from __future__ import annotations

from typing import (Protocol, )
#]


class FromFileableProtocol(Protocol, ):
    """
    """
    #[

    def from_string(source_string: str, /, **kwargs, ) -> Self: ...

    #]


class FromFileMixin:
    """
    """
    #[

    @classmethod
    def from_file(
        klass,
        source_files: str | Iterable[str],
        /,
        **kwargs,
    ) -> Self:
        """
        Create a new object from source file(s)
        """
        source_string = _combine_source_files(source_files, )
        return klass.from_string(source_string, **kwargs, )

    #]


def _combine_source_files(
    source_files: str | Iterable[str],
    /,
    joint: str = "\n\n",
) -> str:
    """
    """
    #[
    if isinstance(source_files, str):
        source_files = [source_files]
    source_strings = []
    for f in source_files:
        with open(f, "r") as fid:
            source_strings.append(fid.read(), )
    return "\n\n".join(source_strings, )
    #]




"""
Data arrays with row and column names
"""


#[
from __future__ import annotations

from typing import (Self, Protocol, )
from numbers import (Number, )
from collections.abc import (Iterable, Iterator, )
import numpy as _np
import dataclasses as _dc
import numpy as _np
import functools as _ft
import itertools as _it

from .series import main as _series
from .databoxes import main as _databoxes
from .plans import main as _plans
from .incidences import main as _incidences
from .conveniences import iterators as _iterators
from . import dates as _dates
#]


__all__ = (
    "HorizontalDataslate",
    "VerticalDataslate",
    "multiple_to_databox",
)


class SlatableProtocol(Protocol, ):
    """
    """
    def get_min_max_shift(self, /, ) -> tuple[int, int]: ...
    def get_databox_names(self, /, ) -> tuple[str, ...]: ...
    def get_fallbacks(self, /, ) -> dict[str, Number]: ...
    def get_overwrites(self, /, ) -> dict[str, Number]: ...
    def get_scalar_names(self, /, ) -> Iterable[str, ...]: ...
    def create_qid_to_logly(self, /, ) -> dict[int, bool]: ...


@_dc.dataclass
class _Dataslate:
    """
    """
    #[

    data: _np.ndarray | None = None
    missing_names: tuple[str, ...] | None = None
    descriptions: Iterable[str] | None = None
    boolex_logly: tuple[bool, ...] | None = None
    _base_indexes: tuple[int, ...] | None = None
    _names: Iterable[str] | None = None
    _dates: tuple[_dates.Dater, ...] | None = None
    _record_shape: tuple[int, int] | None = None
    _min_max_shift: tuple[int, int] | None = None
    _boolex_scalar: tuple[bool, ...] | None = None

    @classmethod
    def from_databox_variant(
        klass,
        databox_variant: _databoxes.Databox | dict,
        names: Iterable[str],
        dates: Iterable[_dates.Dater],
        /,
        *,
        base_indexes: Iterable[int] = None,
        descriptions: dict[str, str | None] | None = None,
        fallbacks: dict[str, Number] | None = None,
        overwrites: dict[str, Number] | None = None,
        scalar_names: Iterable[str] | None = None,
        qid_to_logly: dict[int, bool | None] | None = None,
        min_max_shift: tuple[int, int] = (0, 0),
    ) -> Self:
        """
        Create a dataslate from a databox or dict variant
        """
        self = klass()
        self._names = tuple(names)
        self.missing_names = tuple(_databoxes.Databox.get_missing_names(databox_variant, self._names, ))
        self._dates = dates
        self._base_indexes = base_indexes
        self._populate_data(databox_variant, )
        self._populate_descriptions(descriptions, )
        self._populate_boolex_logly(qid_to_logly, )
        self._fill_fallbacks(fallbacks, )
        self._fill_overwrites(overwrites, )
        self._min_max_shift = min_max_shift
        self._populate_boolex_scalar(scalar_names, )
        return self

    @classmethod
    def iter_variants_from_databox_for_slatable(
        klass,
        slatable: SlatableProtocol,
        databox: _databoxes.Databox | dict,
        base_span: Iterable[_dates.Dater],
        /,
    ) -> Iterator[Self]:
        """
        """
        names = slatable.get_databox_names()
        dates, base_indexes, *min_max_shift = \
            _get_extended_range(slatable, base_span, )
        qid_to_logly = slatable.create_qid_to_logly()
        return klass.iter_variants_from_databox(
            databox, names, dates,
            base_indexes=base_indexes,
            fallbacks=slatable.get_fallbacks(),
            overwrites=slatable.get_overwrites(),
            scalar_names=slatable.get_scalar_names(),
            min_max_shift=min_max_shift,
            qid_to_logly=qid_to_logly,
        )

    @classmethod
    def iter_variants_from_databox(
        klass,
        databox: _databoxes.Databox | dict,
        names: Iterable[str] | None,
        dates: Iterable[_dates.Dater],
        /,
        *,
        fallbacks: dict[str, Number] | None = None,
        overwrites: dict[str, Number] | None = None,
        scalar_names: Iterable[str] | None = None,
        **kwargs,
    ) -> Iterator[Self]:
        """
        """
        names = tuple(names or databox.keys())
        dates = tuple(dates)
        #
        from_to = dates[0], dates[-1] if dates else ()
        item_iterator = \
            _ft.partial(_slate_value_variant_iterator, from_to=from_to, )
        #
        databox_variant_iterator = \
            _databoxes.Databox.iter_variants(databox, item_iterator=item_iterator, names=names, )
        #
        fallbacks_variant_iterator = \
            _databoxes.Databox.iter_variants(fallbacks, item_iterator=item_iterator, names=names, ) \
            if fallbacks else _it.repeat(None, )
        #
        overwrites_variant_iterator = \
            _databoxes.Databox.iter_variants(overwrites, item_iterator=item_iterator, names=names, ) \
            if overwrites else _it.repeat(None, )
        #
        for databox_variant, fallbacks_variant, overwrites_variant \
            in zip(databox_variant_iterator, fallbacks_variant_iterator, overwrites_variant_iterator, ):
            #
            yield klass.from_databox_variant(
                databox_variant, names, dates,
                descriptions=_retrieve_descriptions(databox_variant, names, ),
                fallbacks=fallbacks_variant,
                overwrites=overwrites_variant,
                scalar_names=scalar_names,
                **kwargs,
            )

    @classmethod
    def from_databox_for_slatable(
        klass,
        *args,
        variant: int = 0,
        **kwargs,
    ) -> Self:
        """
        """
        iterator = klass.iter_variants_from_databox_for_slatable(*args, **kwargs, )
        for _ in range(variant, ):
            next(iterator, )
        return next(iterator, )

    @classmethod
    def from_databox(
        klass,
        *args,
        variant: int = 0,
        **kwargs,
    ) -> Self:
        """
        """
        iterator = klass.iter_variants_from_databox(*args, **kwargs, )
        for _ in range(variant, ):
            next(iterator, )
        return next(iterator, )

    @property
    def num_names(self, /, ) -> int:
        """
        """
        return len(self._names)

    @property
    def num_initials(self, /, ) -> int:
        """
        """
        return -self._min_max_shift[0]

    @property
    def num_terminals(self, /, ) -> int:
        """
        """
        return self._min_max_shift[1]

    def remove_initial_data(self, /, ) -> None:
        """
        """
        if self.num_initials:
            self.data = self.data[:, self.num_initials:]

    def _populate_data(
        self,
        databox: _databoxes.Databox | dict,
        /,
        ) -> None:
        """
        Populate slatable data from databox or dict using the first variant
        """
        data = []
        for n in self._names:
            new_data = self.create_nan_vector()
            if n in databox:
                new_data[:] = databox[n]
            data.append(new_data, )
        self.data = self._stack_in(data, )

    def _populate_descriptions(
        self,
        descriptions: dict[str, str] | None,
    ) -> None:
        """
        """
        descriptions = descriptions or {}
        self.descriptions = tuple(
            descriptions.get(n, "") or ""
            for n in self._names
        )

    def _populate_boolex_logly(
        self,
        qid_to_logly: dict[int, bool] | None,
        /,
    ) -> None:
        """
        """
        qid_to_logly = qid_to_logly or {}
        self.boolex_logly = tuple(
            qid_to_logly.get(i, False) or False
            for i in range(len(self._names), )
        )

    def _populate_boolex_scalar(
        self,
        scalar_names: Iterable[str] | None,
        /,
    ) -> None:
        """
        """
        if not scalar_names:
            return
        self._boolex_scalar = tuple(
            n in scalar_names
            for n in self._names
        )

    def _fill_fallbacks(
        self,
        fallbacks: dict[str, Number] | None,
        /,
    ) -> None:
        """
        """
        if not fallbacks:
            return
        for record_id, name in enumerate(self._names, ):
            if name in fallbacks:
                record = self.retrieve_record(record_id, )
                index_nan = _np.isnan(record)
                record[index_nan] = _np.float64(fallbacks[name])
                self.store_record(record, record_id, )

    def _fill_overwrites(
        self,
        overwrites: dict[str, Number] | None,
        /,
    ) -> None:
        """
        """
        if not overwrites:
            return
        for record_id, name in enumerate(self._names, ):
            if name in overwrites:
                record = self.retrieve_record(record_id, )
                record[:] = _np.float64(overwrites[name])
                self.store_record(record, record_id, )

    def to_databox(self, *args, **kwargs, ) -> _dates.Databox:
        """
        """
        return multiple_to_databox((self,), *args, **kwargs, )

    @property
    def num_periods(self, /, ) -> int:
        """
        """
        return len(self._dates)

    @property
    def from_to(self, /, ) -> tuple[_dates.Dater, _dates.Dater]:
        """
        """
        return self._dates[0], self._dates[-1]

    def create_nan_vector(self, /, ) -> int:
        """
        """
        return _np.full((self.num_periods, ), _np.nan, dtype=_np.float64, )

    @property
    def num_rows(self, /, ) -> int:
        """
        """
        return self.data.shape[0] if self.data is not None else 0

    @property
    def base_slice(self, /, ) -> slice:
        """
        """
        return slice(self._base_indexes[0], self._base_indexes[-1]+1)

    def remove_terminal(self, /, ) -> None:
        """
        """
        last_base_index = self._base_indexes[-1]
        self._dates = self._dates[:last_base_index+1]
        self._remove_terminal_data(last_base_index, )

    def copy_data(self, /, ) -> _np.ndarray:
        """
        """
        return self.data.copy()

    def create_name_to_row(
        self,
        /,
    ) -> dict[str, int]:
        return { name: row for row, name in enumerate(self._names, ) }

    @staticmethod
    def retrieve_vector_from_data_array(
        data: _np.ndarray,
        tokens: Iterable[_incidences.Incidence],
        index_zero: int,
        /,
    ) -> _np.ndarray:
        ...

    def retrieve_vector(
        self,
        tokens: tuple[str, ...],
        index_zero: int,
        /,
    ) -> _np.ndarray:
        """
        """
        return self.retrieve_vector_from_data_array(self.data, tokens, index_zero, )

    @staticmethod
    def store_vector_in_data_array(
        vector: _np.ndarray,
        data: _np.ndarray,
        tokens: Iterable[_incidences.Incidence],
        index_zero: int,
        /,
    ) -> None:
        ...

    def store_vector(
        self,
        tokens: tuple[str, ...],
        vector: _np.ndarray,
        index_zero: int,
        /,
    ) -> None:
        """
        """
        store_vector_in_horizontal_data_array(vector, self.data, tokens, index_zero, )

    def retrieve_record(
        self,
        record_id: int,
        /,
        slice_: slice | None = None,
    ) -> _np.ndarray:
        """
        """
        ...

    def store_record(
        self,
        values: _np.ndarray,
        record_id: int,
        /,
        slice_: slice | None = None,
    ) -> None:
        """
        """
        ...

    def _remove_terminal_data(self, /, ) -> None:
        """
        """
        ...

    #]


class HorizontalDataslate(_Dataslate, ):
    """
    """
    #[
    _record_reshape = (1, -1, )

    @staticmethod
    def _stack_in(data, /, ) -> _np.ndarray:
        """
        """
        return _np.vstack(data, )

    @staticmethod
    def retrieve_vector_from_data_array(
        data: _np.ndarray,
        tokens: Iterable[_incidences.Incidence],
        index_zero: int,
        /,
    ) -> _np.ndarray:
        """
        """
        rows, columns = _incidences.rows_and_columns_from_tokens_in_horizontal(tokens, index_zero, )
        return data[rows, columns].reshape(-1, 1)

    @staticmethod
    def store_vector_in_data_array(
        vector: _np.ndarray,
        data: _np.ndarray,
        tokens: Iterable[_incidences.Incidence],
        index_zero: int,
    ) -> None:
        """
        """
        rows, columns = _incidences.rows_and_columns_from_tokens_in_horizontal(tokens, index_zero, )
        data[rows, columns] = vector

    def remove_periods_from_start(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove < 0:
            raise ValueError("Cannot remove negative number of columns from start")
        if remove:
            self.data = self.data[:, remove:]
            self._dates = self._dates[remove:]

    def remove_periods_from_end(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove > 0:
            raise ValueError("Cannot remove positive number of columns from end")
        if remove:
            self.data = self.data[:, :remove]
            self._dates = self._dates[:remove]

    def _remove_terminal_data(
        self,
        last_base_index: int,
        /,
    ) -> None:
        self.data = self.data[:, :last_base_index+1]

    @property
    def column_dates(self, /, ) -> tuple[_dates.Dater, ...]:
        """
        """
        return self._dates

    @property
    def base_columns(self, /, ) -> tuple[int, ...]:
        """
        """
        return self._base_indexes

    @property
    def row_names(self, /, ) -> Iterable[str]:
        """
        """
        return self._names

    def retrieve_record(
        self,
        record_id: int,
        /,
        columns: slice | Iterable[int] | None = None,
    ) -> _np.ndarray:
        """
        """
        if columns is None:
            return self.data[record_id, :]
        else:
            return self.data[record_id, columns]

    def store_record(
        self,
        values: _np.ndarray,
        record_id: int,
        /,
        columns: slice | Iterable[int] | None = None,
    ) -> None:
        """
        """
        if columns is None:
            self.data[record_id, :] = values
        else:
            self.data[record_id, columns] = values

    #]


class VerticalDataslate(_Dataslate, ):
    """
    """
    #[
    _record_reshape = (-1, 1, )

    @staticmethod
    def _stack_in(data, /, ) -> _np.ndarray:
        """
        """
        return _np.hstack(data)

    @property
    def row_dates(self, /, ) -> tuple[_dates.Dater, ...]:
        """
        """
        return self._dates

    @property
    def column_names(self, /, ) -> tuple[str]:
        """
        """
        return self._names

    @property
    def base_rows(self, /, ) -> tuple[int, ...]:
        """
        """
        return self._base_indexes

    def retrieve_record(
        self,
        record_id: int,
        /,
        slice_: slice | None = None,
    ) -> _np.ndarray:
        """
        """
        if slice_ is None:
            return self.data[:, record_id]
        else:
            return self.data[slice_, record_id]

    def _remove_terminal_data(
        self,
        last_base_index: int,
        /,
    ) -> None:
        self.data = self.data[:last_base_index+1, :]

    def store_record(
        self,
        values: _np.ndarray,
        record_id: int,
        /,
        slice_: slice | None = None,
    ) -> None:
        """
        """
        if slice_ is None:
            self.data[:, record_id] = values
        else:
            self.data[slice_, record_id] = values

    #]


def multiple_to_databox(
    dataslates,
    /,
    target_databox: _databoxes.Databox | None = None,
) -> _databoxes.Databox:
    """
    Add data from a dataslate to a new or existing databox
    """
    #[
    if target_databox is None:
        target_databox = _databoxes.Databox()
    num_names = dataslates[0].num_names
    num_variants = len(dataslates)
    start_date = dataslates[0]._dates[0]
    boolex_scalar = dataslates[0]._boolex_scalar or _it.repeat(False, )
    for record_id, name, is_scalar in zip(range(num_names), dataslates[0]._names, boolex_scalar, ):
        data = _np.hstack(tuple(
            ds.retrieve_record(record_id, ).reshape(-1, 1)
            for ds in dataslates
        ))
        if is_scalar:
            value = list(data[0, :])
        else:
            value = _series.Series(
                num_variants=num_variants,
                start_date=start_date,
                values=data,
                description=dataslates[0].descriptions[record_id],
            )
        target_databox[name] = value
    return target_databox
    #]


def _get_extended_range(
    slatable: SlatableProtocol,
    base_span: Iterable[_dates.Dater],
    /,
) -> tuple[Iterable[_dates.Dater], tuple[int, ...]]:
    """
    """
    base_span = tuple(t for t in base_span)
    num_base_periods = len(base_span)
    min_shift, max_shift = slatable.get_min_max_shifts()
    if min_shift == 0:
        min_shift = -1
    min_base_date = min(base_span)
    max_base_date = max(base_span)
    start_date = min_base_date + min_shift
    end_date = max_base_date + max_shift
    base_indexes = tuple(_dates.date_index(base_span, start_date))
    extended_dates = tuple(_dates.Ranger(start_date, end_date))
    return extended_dates, base_indexes, min_shift, max_shift


def _slate_value_variant_iterator(
    value: Any,
    /,
    from_to: tuple[_dates.Dater, _dates.Dater],
) -> Iterator[Any]:
    """
    """
    #[
    if hasattr(value, "iter_data_variants_from_to"):
        return value.iter_data_variants_from_to(from_to, )
    elif isinstance(value, Iterable):
        return _iterators.exhaust_then_last(value, )
    else:
        return _iterators.exhaust_then_last([], value, )
    #]


def _retrieve_descriptions(
    databox: _databoxes.Databox | dict,
    names: Iterable[str],
) -> tuple[str]:
    """
    """
    def _retrieve_item_description(n: str, /, ) -> str:
        if n in databox and hasattr(databox[n], "get_description"):
            return databox[n].get_description()
        else:
            return None
    return { n: _retrieve_item_description(n, ) for n in names }




"""
Sequential models
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, Iterator, )
from typing import (Self, Any, )
import numpy as _np
import copy as _co

from .. import equations as _equations
from .. import quantities as _quantities
from .. import sources as _sources
from ..conveniences import files as _files
from ..incidences import main as _incidences
from ..incidences import blazer as _blazer
from ..explanatories import main as _explanatories
from .. import has_variants as _has_variants

from . import _invariants as _invariants
from . import _variants as _variants
from . import _simulate as _simulate
from . import _assigns as _assigns

#]


__all__ = (
    "Sequential",
)


class Sequential(
    _simulate.SimulateMixin,
    _assigns.AssignMixin,
    _has_variants.HasVariantsMixin,
    _sources.SourceMixin,
    _files.FromFileMixin,
):
    """
    """
    #[

    __slots__ = (
        "_invariant",
        "_variants",
    )

    def __init__(
        self,
        /,
        skeleton: bool = False,
    ) -> None:
        """
        """
        self._invariant = None
        self._variants = []

    def copy(self, /, ) -> Self:
        """
        """
        return _co.deepcopy(self, )

    @classmethod
    def skeleton(
        klass,
        /,
    ) -> Self:
        """
        """
        return klass(skeleton=True, )

    @classmethod
    def from_equations(
        klass,
        equations: Iterable[_equations.Equation],
        /,
        quantities: Iterable[_quantities.Quantity] | None = None,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass()
        self._invariant = _invariants.Invariant.from_equations(equations, quantities, **kwargs, )
        initial_variant = _variants.Variant(self._invariant.parameter_names, )
        self._variants = [ initial_variant ]
        return self

    @classmethod
    def from_source(
        klass,
        source: _sources.ModelSource,
        /,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass.from_equations(
            source.dynamic_equations,
            quantities=source.quantities,
            **kwargs,
        )
        return self

    @property
    def all_names(self, /, ) -> tuple[str]:
        """
        Tuple of names of all variables in order of appearance
        """
        return tuple(self._invariant.all_names)

    @property
    def lhs_names(self, /, ) -> tuple[str]:
        """
        Tuple of names of LHS variables in order of appearance
        """
        return tuple(self._invariant.lhs_names)

    @property
    def residual_names(self, /, ) -> tuple[str]:
        """
        Tuple of names of LHS variables in order of appearance.
        """
        return tuple(self._invariant.residual_names)

    @property
    def rhs_only_names(self, /, ) -> tuple[str]:
        """
        Set of names of RHS variables no appearing on the LHS.
        """
        exclude_names = self._invariant.lhs_names + self._invariant.residual_names + self._invariant.parameter_names
        return tuple(
            n for n in self._invariant.all_names
            if n not in exclude_names
        )

    @property
    def parameter_names(self, /, ) -> tuple[str]:
        """
        Tuple of names of parameters in order of appearance
        """
        return tuple(self._invariant.parameter_names)

    @property
    def identity_index(self, /, ) -> tuple[int]:
        """
        Tuple of indices of identities
        """
        return tuple(
            i for i, x in enumerate(self._invariant.explanatories)
            if x.is_identity
        )

    @property
    def nonidentity_index(self, /, ) -> tuple[int]:
        """
        Tuple of indices of nonidentities
        """
        return tuple(
            i for i, x in enumerate(self._invariant.explanatories)
            if not x.is_identity
        )

    @property
    def equations(self, /, ) -> tuple[_equations.Equation]:
        """
        Tuple of equations in order of appearance
        """
        return tuple(
            x.equation
            for x in self._invariant.explanatories
        )

    @property
    def lhs_quantities(self, /, ) -> tuple[_quantities.Quantity]:
        """
        Tuple of LHS quantities in order of appearance
        """
        lhs_names = self._invariant.lhs_names
        kind = _quantities.QuantityKind.LHS_VARIABLE
        logly = False
        return tuple(
            _quantities.Quantity(qid, name, kind, logly, desc, )
            for (qid, name), desc in zip(enumerate(self._invariant.all_names), self.descriptions)
            if name in lhs_names
        )

    @property
    def num_equations(self, /, ) -> int:
        """
        Number of equations
        """
        return len(self.equations)

    @property
    def descriptions(self, /, ) -> tuple[str]:
        """
        """
        return tuple(
            x.equation.description
            for x in self._invariant.explanatories
        )

    @property
    def incidence_matrix(self, /, ) -> _np.ndarray:
        def _shift_test(tok: _incidences.Token) -> bool:
            return tok.shift == 0
        return _equations.create_incidence_matrix(
            self.equations,
            self.lhs_quantities,
            shift_test=_shift_test,
        )

    @property
    def min_shift(self, /, ) -> int:
        """
        """
        return min(
            x.min_shift
            for x in self._invariant.explanatories
        )

    @property
    def max_shift(self, /, ) -> int:
        """
        """
        return max(
            x.max_shift
            for x in self._invariant.explanatories
        )

    def reorder_equations(self, *args, **kwargs, ) -> None:
        """
        """
        self._invariant.reorder_equations(*args, **kwargs, )

    def sequentialize(
        self,
        /,
    ) -> tuple[int, ...]:
        """
        Reorder the model equations so that they can be solved sequentially
        """
        if self.is_sequential:
            return tuple(range(self.num_equations))
        eids_reordered = _blazer.sequentialize_strictly(self.incidence_matrix, )
        self.reorder_equations(eids_reordered, )
        return tuple(eids_reordered)

    def iter_explanatories(self, /, ) -> Iterator[_explanatories.Explanatory]:
        """
        """
        yield from self._invariant.explanatories

    @property
    def is_sequential(
        self,
        /,
    ) -> bool:
        """
        """
        return _blazer.is_sequential(self.incidence_matrix, )

    def get_description(self, /, ) -> str:
        """
        """
        return self._invariant.get_description()

    def set_description(self, *args, **kwargs, ) -> None:
        """
        """
        self._invariant.set_description(*args, **kwargs, )

    def set_extra_databox_names(self, *args, **kwargs, ) -> None:
        """
        """
        self._invariant.set_extra_databox_names(*args, **kwargs, )

    def get_human_equations(
        self,
        /,
        descriptions: bool = True,
        separator: str = "\n\n",
    ) -> tuple[str]:
        """
        """
        return tuple(
            x.equation.human
            for x in self._invariant.explanatories
        )

    def create_qid_to_name(self, *args, **kwargs, ) -> dict[int, str]:
        """
        """
        return self._invariant.create_qid_to_name(*args, **kwargs, )

    def create_name_to_qid(self, *args, **kwargs, ) -> dict[str, int]:
        """
        """
        return self._invariant.create_name_to_qid(*args, **kwargs, )

    def __repr__(self, /, ) -> str:
        """
        """
        indented = " " * 4
        return "\n".join((
            f"",
            f"{self.__class__.__name__} model",
            f"Description: \"{self.get_description()}\"",
            f"|",
            f"| Num of variants: {self.num_variants}",
            f"| Num of equations: {self.num_equations}",
            f"| Num of [nonidentities, identities]: [{len(self.nonidentity_index)}, {len(self.identity_index)}]",
            f"| Num of parameters: {len(self.parameter_names)}",
            f"| Num of rhs-only variables: {len(self.rhs_only_names)}",
            f"| Time shifts [min, max]: [{self.min_shift:+g}, {self.max_shift:+g}]",
            f"|",
        ))

    def __str__(self, /, ) -> str:
        """
        """
        return repr(self, )

    def __getitem__(
        self,
        request: int,
        /,
    ) -> Self:
        """
        """
        return self.get_variant(request, )

    #
    # ===== Implement SlatableProtocol =====
    #

    def get_min_max_shifts(self, /, ) -> tuple[int, int]:
        """
        """
        return self.min_shift, self.max_shift

    def get_databox_names(self, /, ) -> tuple[str]:
        """
        """
        extra_databox_names = self._invariant.extra_databox_names or ()
        return tuple(self._invariant.all_names) + tuple(extra_databox_names)

    def get_fallbacks(self, /, ) -> dict[str, Any]:
        """
        """
        return { n: 0 for n in tuple(self._invariant.residual_names) }

    def get_overwrites(self, /, ) -> dict[str, Any]:
        """
        """
        return self.get_parameters()

    def get_scalar_names(self, /, ) -> tuple[str]:
        """
        """
        return tuple(self.parameter_names, )

    def create_qid_to_logly(self, /, ) -> dict[str, bool]:
        """
        """
        return {}

    #
    # ===== Implement PlannableSimulateProtocol =====
    #

    @property
    def simulate_can_be_exogenized(self, /, ) -> tuple[str, ...]:
        return tuple(set(
            i.lhs_name for i in self._invariant.explanatories
            if not i.is_identity
        ))

    @property
    def simulate_can_be_endogenized(self, /, ) -> tuple[str, ...]:
        return tuple(set(
            i.residual_name for i in self._invariant.explanatories
            if not i.is_identity
        ))

    simulate_can_be_anticipated = ()
    simulate_can_be_when_data = simulate_can_be_exogenized

    #]





"""
"""


#[
from __future__ import absolute_import

from collections.abc import (Iterable, )
from typing import (Protocol, )
#]


class AttributesProtocol(Protocol, ):
    """
    """
    #[

    def has_attributes(self, *attributes) -> bool:  ...
    attributes: set[str]

    #]


class AttributesMixin:
    """
    """
    #[

    def has_attributes(self, *attributes) -> bool:
        attributes = (set((a, )) if isinstance(a, str) else set(a) for a in attributes)
        return all(self.attributes.intersection(a, ) for a in attributes)

    #]


def generate_by_attributes(
    with_attributes: Iterable[AttributesProtocol],
    *args,
) -> Iterable[AttributesProtocol]:
    """
    """
    return ( i for i in with_attributes if i.has_attributes(*args, ) )




"""
First-order system simulators
"""


#[
from __future__ import annotations

from typing import (Self, Any, TypeAlias, Literal, Protocol, runtime_checkable)
from collections.abc import (Iterable, )
import numpy as _np

from ..databoxes import main as _databoxes
from ..fords import simulators as _ford_simulators
from ..fords import solutions as _solutions
from ..periods import simulators as _period_simulators
from ..plans import main as _plans
from .. import dataslates as _dataslates

from . import main as _simultaneous
#]


_DATASLATE_VARIANT_ITERATOR = \
    _dataslates.HorizontalDataslate.iter_variants_from_databox_for_slatable


class SimulateMixin:
    """
    """
    #[

    def simulate(
        self,
        in_databox: _databoxes.Databox,
        span: Iterable[Dater],
        /,
        *,
        plan: _plans.PlanSimulate | None = None,
        method: Literal["first_order", "period", "stacked"] = "first_order",
        prepend_input: bool = True,
        target_databox: _databoxes.Databox | None = None,
        num_variants: int | None = None,
        **kwargs,
    ) -> tuple[_databoxes.Databox, dict[str, Any]]:
        """
        """
        num_variants = (
            num_variants
            if num_variants is not None
            else self.num_variants
        )
        base_dates = tuple(span, )
        if plan is not None:
            plan.check_consistency(self, base_dates, )
        #
        out_dataslates = []
        zipped = zip(
            range(num_variants, ),
            self.iter_variants(),
            _DATASLATE_VARIANT_ITERATOR(self, in_databox, base_dates, )
        )
        #
        #=======================================================================
        # Main loop over variants
        for vid, mdi, dsi in zipped:
            #
            # Simulate and write to dataslate
            _SIMULATE_METHODS[method](
                mdi, dsi, vid,
                plan=plan,
                **kwargs,
            )
            dsi.remove_terminal()
            out_dataslates.append(dsi, )
        #=======================================================================
        #
        # Combine resulting dataslates into a databox
        out_db = _dataslates.multiple_to_databox(out_dataslates, )
        if prepend_input:
            out_db.prepend(in_databox, base_dates[0]-1, )
        out_db = out_db | self.get_parameters_stds()
        #
        # Add to custom databox
        if target_databox is not None:
            out_db = target_databox | out_db
        #
        # Simulation info
        info = {}
        #
        return out_db, info

    #]


def _simulate_first_order(
    model: _simultaneous.Simultaneous,
    dataslate: _dataslates.Dataslate,
    vid: int,
    **kwargs,
) -> None:
    """
    """
    _ford_simulators.simulate_flat(
        model._variants[0].solution,
        model.get_solution_vectors(),
        dataslate,
        vid,
        **kwargs,
    )


_SIMULATE_METHODS = {
    "first_order": _simulate_first_order,
    "period": _period_simulators.simulate,
}


